
/*!
 * \file
 * \author Toni Corvera
 *
 * $Id: AutoIndex.cc 483 2013-06-17 17:41:04Z toni $
 */
/*
 * Copyright (c) 2012 Toni Corvera
 *
 * This file is part of TFCWeb.
 *
 * TFCWeb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TFCWeb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TFCWeb. If not, see <http://www.gnu.org/licenses/>.
 */

#include "AutoIndex.h"
#include "Servidor.h"
#include "TipusMIME.h"
#include "utils.h"

#include <algorithm>
#include <sstream>
#include <vector>
#include <boost/date_time.hpp>
#include <boost/lexical_cast.hpp>

using namespace boost::filesystem;
using namespace std;
using namespace tfc::utils;

using boost::lexical_cast;

namespace MIME = tfc::RegistreTipusMIME;

namespace {

//! \brief Amplades (en caracters) dels elements de l'índex
enum {
    AMPLE_DATA = 11, //!< Amplada de la data de modificació del fitxer
    AMPLE_MIDA = 12, //!< Amplada de la mida del fitxer
    AMPLE_MIME = 15  //!< Amplada del tipus MIME
};

// Drecera
inline string mtime(const time_t & t) {
    return tfc::utils::formata_data(t, "%d %b %Y");
}

string mida(const size_t s) {
    if (s > (1<<20)) {
        const size_t mb = s / (1<<20);
        const size_t kb = s % (1<<20);
        return lexical_cast<string>(mb)+"."+lexical_cast<string>(kb)[0]+" MiB";
    }
    else if (s > (1<<10)) {
        const size_t kb = s / (1<<10);
        const size_t b = s % (1<<10);
        return lexical_cast<string>(kb)+"."+lexical_cast<string>(b)[0]+" KiB";
    }
    return lexical_cast<string>(s)+" B";
}

} // ns anònim

namespace tfc {

string AutoIndex::html() const {
    if (!cache_.empty()) { // Si ja s'ha generat no cal repetir-ho
        return cache_;
    }

    stringstream ss;

    const bool barra_final = ( !url_.ruta().empty() && '/' == utils::final(url_.ruta()));
    const string base_dir = url_.ruta() + ( barra_final ? "" : "/");
    /*
     * Formats:
     * Apache:
     *   Llistat de noms enllaçats
     * Apache FancyIndex:
     *   Name, Last Modified, Size, Description
     * Webfsd:
     *   access, user, group, date, size, name
     */
    // FIXME: Abstraure?
    ss << "<!DOCTYPE html>\n"
          "<html>\n"
          "<head><title>Índex per a " << base_dir << "</title></head>\n"
       << "<body>\n"
          "<h1>Índex per a " << base_dir << "</h1>\n"
          "<pre>"
       << lpad("data", AMPLE_DATA)
       << rpad("  tipus", AMPLE_MIME)
       << lpad("mida", AMPLE_MIDA)
       << "   nom"
          "</pre>"
          "<hr />\n"
          "<pre>";
    // El llistat de directoris obté els fitxers desordenats, cal un pas intermig per ordenar-los
    // <http://www.boost.org/doc/libs/1_49_0/libs/filesystem/v3/doc/tutorial.html#Using-path-decomposition>
    vector<path> paths;

    //paths.push_back(path("."));
    paths.push_back(path(".."));
    copy(directory_iterator(directori()), directory_iterator(), back_inserter(paths));
    sort(paths.begin(), paths.end());

    //const string base = "http://" + to_string(host_);

    const string str_dir = string(AMPLE_MIDA-5, ' ')+"&lt;DIR&gt;";
    // FIXME: Ignorar dotfiles i il·legibles (Permisos? -Windows: attrib-)
    for (auto it = paths.begin(); it != paths.end(); ++it) {
        if (!es_llegible(*it)) {
            continue;
        }
        // La conversió implícita a string afegeix cometes
        const string fn = it->filename().string();
        const bool es_dir = is_directory(*it);
        const string str_mida = (es_dir ? str_dir : mida(file_size(*it)));
        const string data = mtime(last_write_time(*it));
        const string extra = (es_dir ? "/" : "");
        const string mime = MIME::tipus(*it);
        const string url = base_dir + fn;
        ss
           << lpad(data, AMPLE_DATA)
           << rpad("  "+mime, AMPLE_MIME)
           << lpad(str_mida, AMPLE_MIDA)
           << "    <a href=\"" << url << "\">" // FIXME: Construir path relatiu per enllaç
           << fn << extra << "</a>"
           << "\n";
    }

    ss << "</pre>\n"
          "<hr />\n"
          "<address>" << Servidor::signatura()
       << ". Servidor " << host_.address() << ", port " << host_.port()
       << "</address>\n"
          "</body>\n"
          "</html>";
    cache_ = ss.str();
    return cache_;
}

void AutoIndex::envia(boost::asio::ip::tcp::socket & socket) const
    throw (runtime_error)
{
    this->envia(socket, html());
}

AutoIndex::tipus_capsaleres AutoIndex::capsaleres() const {
    tipus_capsaleres capsaleres = RecursDinamic::capsaleres();
    const string content_length = lexical_cast<string>(html().length());
    capsaleres.push_back( make_pair("Content-Length", content_length) );
    // FIXME: Accept-Charset?? vs UTF-8
    // TODO: espai entre MIME i charset?
    capsaleres.push_back( make_pair("Content-Type", "text/html; charset=UTF-8") );
    // TODO: Afegir Last-Modified?
    return capsaleres;
}

} // ns tfc

// vim:set ts=4 et ai: //
