#if !defined(_CONFIGURACIO_H_)
#define _CONFIGURACIO_H_

/*!
 * \file
 * \author Toni Corvera
 *
 * $Id: Configuracio.h 483 2013-06-17 17:41:04Z toni $
 */
/*
 * Copyright (c) 2012 Toni Corvera
 *
 * This file is part of TFCWeb.
 *
 * TFCWeb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TFCWeb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TFCWeb. If not, see <http://www.gnu.org/licenses/>.
 */

#include "Excepcions.h"
#include "Permisos.h"
#include "portabilitat.h"
#include "utils.h"

#include <ostream>
#include <string>
#include <boost/filesystem.hpp>

namespace tfc {

/*!
 * \brief Classe base dels error en les opcions (configuració o línia de comandes)
 */
class ErrorOpcions : public ErrorTFC {
protected:
    explicit ErrorOpcions(const std::string & w)
        : ErrorTFC(w)
    {
        // buit
    }
public:
    virtual unsigned int codi() const = 0;
};

/*!
 * \brief Representa un error en la configuració
 */
class ErrorConfiguracio : public ErrorOpcions {
public:
    explicit ErrorConfiguracio(const std::string & w = "Error en la configuració")
        : ErrorOpcions(w)
    {
        // buit
    }
    virtual unsigned int codi() const { return EX_CONFIG; }
};

/*!
 * \brief Representa un error en la línia de comandes
 */
class ErrorLiniaComandes : public ErrorOpcions {
public:
    explicit ErrorLiniaComandes(const std::string & w = "Error en la línia de comandes")
        : ErrorOpcions(w)
    {
        // buit
    }
    virtual unsigned int codi() const { return EX_USAGE; }
};

/*!
 * \brief Indica que s'ha demanat ajuda a la línia de comandes
 *        ($ tfcweb --help)
 * \note La convenció és sortir immediatament un cop mostrada
 */
class AjudaDemanada : public ErrorTFC {
public:
    explicit AjudaDemanada(const std::string & ajuda)
        : ErrorTFC(ajuda)
    {
        // buit
    }
};

/*!
 * \brief Indica que s'ha demanat informació de versió a la línia
 *        de comandes ($ tfcweb --version)
 * \note La convenció és sortir immediatament un cop mostrada
 */
class VersioDemanada : public ErrorTFC {
public:
    VersioDemanada()
        : ErrorTFC("")
    {
        // buit
    }
};

/*!
 * \brief Configuració carregada
 * \invariant \c arrel() existeix, és un directori i és canònica
 */
class Configuracio {
public:
    /*!
     * \brief Llegeix la configuració i la línia de comandes
     * \param argc Nombre d'arguments de la línia de comandes
     * \param argv Arguments de la línia de comandes
     * \throws AjudaDemanada Si l'usuari ha demanat l'ajuda
     * \throws VersioDemanada Si l'usuari ha demanat la informació de versions
     * \throws ErrorLiniaComandes Si la línia de comandes conté errors
     * \throws ErrorConfiguracio Si l'arxiu de configuració existeix i conté errors
     * \post La línia de comandes s'ha processat sense errors
     * \post La configuració s'ha llegit correctament si existeix
     */
    Configuracio(int argc, char * argv[]) throw (ErrorOpcions,AjudaDemanada,VersioDemanada);
    Configuracio(const Configuracio & c)
        : argc_(c.argc_), argv_(c.argv_),
          arrel_(c.arrel_), index_(c.index_), port_(c.port_), config_(c.config_),
          auth_user_(c.auth_user_), auth_pass_(c.auth_pass_),
          cerber_(c.cerber_)
    {
        // buit
    }

    const boost::filesystem::path& arrel() const { return arrel_; }
    const std::string& index() const { return index_; }
    unsigned short port() const { return port_; }
    const boost::filesystem::path& config() const { return config_; }
    const std::string& usuari() const { return auth_user_; }
    const std::string& contrasenya() const { return auth_pass_; }
    const size_t fils() const { return fils_; }
    const std::string & realme() const { return auth_realme_; }

    const Cerber& cerber() const { return cerber_; }
protected:
    // C++11: Configuracio& operator=(const Configuracio &) = delete;
    Configuracio& operator=(const Configuracio &) {
        throw ErrorNoImplementat(FITXER_I_LINIA_);
        return *this;
    }
private:
    const int argc_;
    char ** argv_;
    /*!
     * \invariant Ruta absoluta
     */
    boost::filesystem::path arrel_;
    std::string index_;
    unsigned short port_;
    boost::filesystem::path config_;
    std::string auth_user_, auth_pass_, auth_realme_;
    size_t fils_;
    Cerber cerber_;

    //! \brief Impressió de configuració
    friend std::ostream& operator<<(std::ostream & os, const Configuracio & c);
};

} // namespace tfc

#endif // _CONFIGURACIO_H_

// vim:set ts=4 et ai: //
