#if !defined(_COS_RESPOSTA_H_)
#define _COS_RESPOSTA_H_

/*!
 * \file
 * \author Toni Corvera
 *
 * $Id: CosResposta.h 483 2013-06-17 17:41:04Z toni $
 */
/*
 * Copyright (c) 2012 Toni Corvera
 *
 * This file is part of TFCWeb.
 *
 * TFCWeb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TFCWeb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TFCWeb. If not, see <http://www.gnu.org/licenses/>.
 */

#include "CodisEstatHTTP.h"
#include "Excepcions.h"
#include "Servidor.h"
#include "Transformacions.h"

#include <memory>
#include <string>
#include <vector>
#include <boost/asio.hpp>

namespace tfc {

#if 0
/*!
 * \brief Tipus de codificacions de contingut conegudes, per aplicar a les respostes
 * \remarks Correspondència amb la capçalera \c Content-Encoding. És una propietat
 *          de l'entitat original, generalment indica que el recurs està emmagatzemat
 *          ja amb aquesta codificació, mentre que quan la codificació es fa al
 *          vol es correspon al \c Transfer-Encoding (\ref CodificacioTransmissio)
 * \remarks No s'implementa cap codificació (excepte \c CE_IDENTITY que és l'identitat)
 * \sa http://tools.ietf.org/html/rfc2616#section-3.5
 */
enum ContentEncoding {
    CE_CAP = 0,
    CE_IDENTITY,
    CE_GZIP,
    CE_COMPRESS,
    CE_DEFLATE
};

/*!
 * \brief Tipus de codificacions de transmissió conegudes, per aplicar a les respostes
 * \remarks Correspondència amb la capçalera \c Transfer-Encoding. És una propietat
 *          del missatge
 * \remarks Només s'implementa \em chunked
 * \remarks L'implementació de \em chunked és obligatòria, totes les aplicacions
 *          han de ser capaces de rebre missatges \em chunked (RFC 2616, pàg. 26)
 * \sa http://tools.ietf.org/html/rfc2616#section-3.6
 */
enum TransferEncoding {
    TE_CAP = 0,
    TE_IDENTITY,
    TE_CHUNKED,
    TE_GZIP,
    TE_COMPRESS,
    TE_DEFLATE
};
#endif

/*!
 * \brief Classe base del cos dels missatges de resposta
 */
class CosResposta {
public:
    typedef std::pair<std::string, std::string> tipus_capsalera;
    typedef std::vector<tipus_capsalera> tipus_capsaleres;
    typedef std::shared_ptr<Transformacio> transformacio_ptr;
    typedef boost::asio::ip::tcp::socket socket_tcp;

    /*!
     * \param peticio La petició HTTP enviada pel client
     */
    explicit CosResposta(const PeticioHTTP & peticio)
        : peticio_(peticio),
          deflate_ok_(false), chunked_ok_(false),
          trans_()
    {
        determina_codificacions();
    }

    CosResposta(const CosResposta & cr)
        : peticio_(cr.peticio_),
          deflate_ok_(cr.deflate_ok_),
          chunked_ok_(cr.chunked_ok_),
          te_(cr.te_), ce_(cr.ce_),
          trans_(cr.trans_)
    {

    }

    CosResposta& operator=(const CosResposta & cr) {
        peticio_ = cr.peticio_;
        deflate_ok_ = cr.deflate_ok_;
        chunked_ok_ = cr.chunked_ok_;
        te_ = cr.te_;
        ce_ = cr.ce_;
        trans_ = cr.trans_;
        return *this;
    }

    virtual ~CosResposta() {
        // buit
    }

    /*!
     * \brief Envia el cos del missatge
     * \todo XXX: Utilitzar algun tipus de "sink" potser seria preferible
     * \remarks Cal cridar envia_eom() després
     */
    virtual void envia(socket_tcp & socket) const
                                throw (std::runtime_error) = 0;

    /*!
     * \brief Envia el final del missatge.
     * \remarks Quan s'emprea una codificació cal finalitzar els missatges.
     */
    virtual void envia_eom(socket_tcp & socket) const throw (ErrorEscrivintEnSocket);

    /*!
     * \brief Llista de capçaleres específiques del cos a afegir a la resposta.
     * Per defecte, retorna les capçaleres relatives a les transformacions
     * aplicables.
     */
    virtual tipus_capsaleres capsaleres() const;

    /*!
     * \brief Indica si el contingut pot generar capçaleres HTTP.
     * Per a alguns tipus de recursos dinàmics, com ara scripts, és
     * dessitjable poder modificar les capçaleres.
     * \remarks Un CosResposta que generi capçaleres haurà d'enviar
     *          opcionalment les capçaleres generades, seguides d'un
     *          "CRLF" abans del contingut pròpiament dit per a
     *          marcar el final de les capçaleres.
     * \remarks Les capçaleres retornades amb \ref capsaleres()
     *          les afegirà el \ref Servidor. L'ús de \ref pot_generar_capsaleres()
     *          no hi està relacionat.
     */
    virtual bool pot_generar_capsaleres() const {
        return false;
    }

    const PeticioHTTP& peticio() const { return peticio_; }

protected:
    /*!
     * \brief Envia una cadena sencera
     * \param socket Socket on escriure
     * \param buffer Cadena a escriure
     * \pre Socket connectat FIXME: <- Cal? Comprovar?
     * \throws ErrorEscrivintEnSocket si falla l'enviament de dades
     * \remarks S'utilitza la codificació definida
     */
    virtual size_t envia(socket_tcp & socket,
                         const std::string & buffer) const throw (ErrorEscrivintEnSocket);

    /*!
     * \brief Envia una cadena sense aplicar la codificació
     */
    size_t envia_literal(socket_tcp & socket, const std::string & buffer)
                                                const throw (ErrorEscrivintEnSocket);


    //! Comprova si el client accepta \em deflate
    bool deflate_ok() const { return deflate_ok_; }

    //! Comprova si el client accepta \em chunked
    bool chunked_ok() const { return chunked_ok_; }

    /*!
     * \brief Genera una transformació acceptable per el client
     */
    transformacio_ptr transformacio() const { return trans_; }

    //! \c Content-Encoding acceptat per el client
    const std::string& content_encoding() const { return ce_; }

    //! \c Transfer-Endocing acceptat per el client
    const std::string& transfer_encoding() const { return te_; }
private:
    //! Defineix una codificació de contingut/transferència apropiada al client
    void determina_codificacions();

    PeticioHTTP peticio_; //!< Petició associada
    bool deflate_ok_; //!< El client accepta \em deflate
    bool chunked_ok_; //!< El client accepta \em chunked
    std::string te_, //!< \c Transfer-Encoding
                ce_; //!< \c Content-Encoding
    transformacio_ptr trans_; //!< Transformació
};

class RecursDinamic : public CosResposta { // FIXME: Implementació temporal
public:
    explicit RecursDinamic(const PeticioHTTP & p)
        : CosResposta(p)
    {
        // buit
    }
    RecursDinamic(const RecursDinamic & rd)
        : CosResposta(rd)
    {
        // buit
    }
    RecursDinamic& operator=(const RecursDinamic & rd) {
        CosResposta::operator=(rd);
        return *this;
    }

    /*!
     * \brief Contingut
     * \remarks Per evitar col·locar tot en memòria cal sobrecarregar \ref envia()
     */
    virtual std::string html() const = 0;

    virtual void envia(socket_tcp & socket) const
                        throw (std::runtime_error)
    {
        this->envia(socket, html());
    }
protected:
    using CosResposta::envia; // envia(socket, string)
};

class CosPaginaError : public RecursDinamic {
public:
    CosPaginaError(const PeticioHTTP & p, CodiEstatHTTP error, const std::string & missatge)
        : RecursDinamic(p), error_(error), missatge_(missatge), cache_()
    {
        // buit
    }
    CosPaginaError(const CosPaginaError & cpe)
        : RecursDinamic(cpe), error_(cpe.error_), missatge_(cpe.missatge_), cache_(cpe.cache_)
    {
        // buit
    }

    virtual tipus_capsaleres capsaleres() const;
    virtual std::string html() const;
private:
    CosPaginaError& operator=(const CosPaginaError & cpe) {
        throw ErrorNoImplementat(FITXER_I_LINIA_);
    }
    const CodiEstatHTTP error_;
    const std::string missatge_;
    mutable std::string cache_;
};

/*!
 * \brief Creació d'una resposta a partir d'un text
 */
class CosTextual : public RecursDinamic {
public:
    CosTextual(const PeticioHTTP & p, const std::string & text)
        : RecursDinamic(p), text_(text)
    {
        // buit
    }
    CosTextual(const CosTextual & c)
        : RecursDinamic(c), text_(c.text_)
    {
        // buit
    }

    virtual string html() const { return text_; }
private:
    CosTextual& operator=(const CosTextual & c) {
        throw ErrorNoImplementat(FITXER_I_LINIA_);
    }
    const std::string text_;
};

} // ns tfc

#endif // _COS_RESPOSTA_H_

// vim:set ts=4 et ai: //
