#if !defined(_CUA_CONCURRENT_H_)
#define _CUA_CONCURRENT_H_

/*!
 * \file
 * \author Toni Corvera
 *
 * $Id: CuaConcurrent.h 483 2013-06-17 17:41:04Z toni $
 */
/*
 * Copyright (c) 2012 Toni Corvera
 *
 * This file is part of TFCWeb.
 *
 * TFCWeb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TFCWeb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TFCWeb. If not, see <http://www.gnu.org/licenses/>.
 */

#include "portabilitat.h"

#include <algorithm>
#include <deque>
#include <boost/thread.hpp>

namespace tfc {

// Nota: boost::unique_lock<boost::mutex> i boost::mutex::scoped_lock són equivalents

// A diferència de la majoria del projecte es fan servir noms en anglès per
// apropar-se a la STL
// Altres fonts d'informació:
//  - Implementació de cua concurrent sense locks: http://www.drdobbs.com/parallel/211601363
/*!
 * \brief Implementació d'una cua concurrent amb \em locking
 * \tparam T Tipus dels elements de la cua
 * \tparam TContenidor Tipus del contenidor utilitzat internament
 * \remarks Interfície basada en \c std::queue. Com \c std::queue es fa
 *          servir el patró decorador.
 * \sa Implementació d'una cua concurrent:
 *     http://www.justsoftwaresolutions.co.uk/threading/implementing-a-thread-safe-queue-using-condition-variables.html
 * \sa Proposta "C++ Concurrent Queues":
 *     http://www.open-std.org/jtc1/sc22/wg21/docs/papers/2012/n3353.html
 */
template<typename T, typename TContenidor = std::deque<T> >
class CuaConcurrent {
private:
    // Drecera
    typedef boost::unique_lock<boost::mutex> scope_lock;
public:
    // Tipus definits per els contenidors de la STL (només els que es fan servir)
    typedef T value_type; //!< Tipus del objectes a la cua
    typedef T & reference; //!< Referència a \c value_type
    typedef const T & const_reference; //!< Referència constant a \c value_type
    typedef typename TContenidor::size_type size_type; //!< Tipus de la mida de la cua
    //! Tipus de les marques de temps en milisegons
    typedef boost::posix_time::milliseconds milliseconds; // Drecera

    //! \brief Constructor per defecte
    CuaConcurrent() {
        // buit
    }

    //! \brief Constructor de còpia
    CuaConcurrent(const CuaConcurrent & c)
    {
        //impl_.reserve(c.impl_.size());
        scope_lock lock(c.mtx_);
        // No cal blocar `this`
        std::copy(c.impl_.begin(), c.impl_.end(), impl_.begin());
    }

    virtual ~CuaConcurrent() {}

    /*!
     * \brief Extreu el primer objecte amb una espera màxima determinada
     * \param[out] desti Valor extret
     * \param[in] ms Temps d'espera màxim
     * \remarks Bloca un màxim de \p ms milisegons
     * \returns \c true si s'ha extret un objecte, \c false si s'ha exhaurit el
     *          temps
     * \throws boost::thread_interrupted Si el fil s'interromp mentre s'esperava
     * \throws boost::thread_exception Si es produeix un altre tipus d'error
     * \sa pop(reference)
     */
    bool pop(reference desti, const milliseconds & ms)
        throw (boost::thread_exception, boost::thread_interrupted)
    {
        // Seria millor utilitzar scoped_timed_lock?
        scope_lock lock(mtx_);
        while (impl_.empty()) {
            // Esperar notificació, desbloca mutex fins que retorni
            const bool ok = cond_.timed_wait(lock, ms);
            if (!ok) {
                return false;
            }
        }
        desti = impl_.front();
        impl_.pop_front();
        return true;
    }

    /*!
     * \brief Extreu el primer objecte
     * \param[out] desti Valor extret
     * \remarks Bloca fins que hi hagi algun element
     * \sa pop(reference,milliseconds)
     * \throws boost::thread_interrupted Si el fil s'interromp mentre s'esperava
     * \throws boost::thread_exception Si es produeix un altre tipus d'error
     */
    void pop(reference desti)
        throw (boost::thread_exception, boost::thread_interrupted)
    {
        scope_lock lock(mtx_);
        while (impl_.empty()) {
            cond_.wait(lock); // Esperar notificació, desbloca mutex fins que retorni
        }
        boost::posix_time::millisec(1);
        desti = impl_.front();
        impl_.pop_front();
    }

    /*!
     * \brief Encua un objecte
     * \param element Element a encuar
     * \remarks Bloca
     */
    void push(const_reference element) {
        scope_lock lock(mtx_);
        const bool cal_notificar = impl_.empty();
        impl_.push_back(element);
        // alliberar assegura que el thread notificat pot accedir immediatament
        lock.unlock();
        if (cal_notificar) {
            cond_.notify_one();
        }
    }

    /*!
     * \brief Extreu de manera immediata o falla
     *
     * Si la cua no està buida ni blocada s'extreu un objecte de manera immediata.
     * \param[out] desti Valor extret
     * \returns \c true si s'ha pogut extreure un objecte, \c false si no
     */
    bool try_pop(reference desti) {
        boost::try_mutex::scoped_try_lock lock(mtx_);
        if (!lock || impl_.empty()) {
            return false;
        }
        desti = impl_.front();
        impl_.pop_front();
        return true;
    }

    /*!
     * \brief Encua un objecte de manera immediata o falla
     *
     * Si la cua està blocada no s'encuarà
     * \param element Objecte a encuar
     * \returns \c true si s'ha pogut encuar, \c false si no
     */
    bool try_push(const_reference element) {
        boost::try_mutex::scoped_try_lock lock(mtx_);
        if (!lock) {
            return false;
        }
        impl_.push_back(element);
        return true;
    }

    /*!
     * \brief Comprova si la cua està buida
     * \remarks Bloca
     */
    bool empty() const {
        boost::mutex::scoped_lock lock(mtx_);
        return impl_.empty();
    }

    /*!
     * \brief Obté la mida de la cua (nombre d'elements encuats)
     * \remarks Bloca
     */
    size_type size() const {
        boost::mutex::scoped_lock lock(mtx_);
        return impl_.size();
    }

    // Degut a les "race conditions" no és recomanable utilitzar front() i back()
#if 0
    /*!
     * \brief Accedeix al primer objecte
     * \pre La cua no és buida
     */
    reference front() {
        UNUSED boost::mutex::scoped_lock lock(mtx_);
        return impl_.front();
    }

    //! \copydoc front()
    const_reference front() const {
        // Cal blocar?
        UNUSED boost::mutex::scoped_lock lock(mtx_);
        return impl_.front();
    }

    /*!
     * \brief Accedeix al darrer objecte
     * \pre La cua no és buida
     */
    const_reference back() const {
        UNUSED boost::mutex::scoped_lock lock(mtx_);
        return impl_.back();
    }

    //! \copydoc back()
    reference back() {
        UNUSED boost::mutex::scoped_lock lock(mtx_);
        return impl_.back();
    }
#endif
private:
    TContenidor impl_; //!< Contenidor real
    boost::condition_variable cond_; //!< Variable de condició usada en l'extracció
    mutable boost::mutex mtx_; //!< Mutex usat en els bloquejos
};

} // ns tfc

#endif // _CUA_CONCURRENT_H_

// vim:set ts=4 et ai: //
