#if !defined(_EXCEPCIONS_H_)
#define _EXCEPCIONS_H_

/*!
 * \file
 * \brief Excepcions base
 * \remarks Totes les excepcions definides en aquest projecte deriven de
 *          \c ErrorTFC i fan servir noms que començen per \c Error. A més
 *          totes permeten la construcció sense paràmetres.
 * \remarks En aquest fitxer es defineixen les excepcions més genèriques,
 *          cada mòdul defineix les corresponents a partir d'aquestes.
 * \author Toni Corvera
 *
 * $Id: Excepcions.h 483 2013-06-17 17:41:04Z toni $
 */
/*
 * Copyright (c) 2012 Toni Corvera
 *
 * This file is part of TFCWeb.
 *
 * TFCWeb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TFCWeb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TFCWeb. If not, see <http://www.gnu.org/licenses/>.
 */

#include "CodisEstatHTTP.h"
#include "portabilitat.h"

#include <stdexcept>
#include <string>
#include <boost/lexical_cast.hpp>

/*
 * http://www.cplusplus.com/reference/std/stdexcept/
 *
 * Logic errors:
 * logic_error      -> Logic error exception
 * domain_error     -> Domain error exception
 * invalid_argument -> Invalid argument exception
 * length_error     -> Length error exception
 * out_of_range     -> Out-of-range exception
 *
 * Runtime errors:
 * runtime_error    -> Runtime error exception
 * range_error      -> Range error exception
 * overflow_error   -> Overflow error exception
 * underflow_error  -> Underflow error exception
 */

namespace tfc {

/*!
 * \brief Excepció base de totes les excepcions del projecte
 */
class ErrorTFC : public std::runtime_error {
public:
    explicit ErrorTFC(const std::string & w = "") throw ()
        : std::runtime_error(w)
    {
        // buit
    }
};

/*!
 * \brief Indica que s'ha intentat l'accés a una ruta d'arxiu inexistent
 */
class ErrorRutaInexistent : public ErrorTFC {
public:
    ErrorRutaInexistent(const std::string & w = "Ruta inexistent")
        : ErrorTFC(w)
    {
        // buit
    }
};

class ErrorEscrivintEnSocket : public ErrorTFC {
public:
    explicit ErrorEscrivintEnSocket(const std::string & w = "Error enviant dades")
        : ErrorTFC(w)
    {
        // buit
    }
};

/*!
 * \brief Error base HTTP; normalment indica que es respondrà al
 *        client amb un missatge d'error.
 */
class ErrorHTTP : public ErrorTFC {
public:
    /*!
     * \param w Descripció de l'error
     * \param c Codi HTTP de l'error
     */
    ErrorHTTP(const std::string & w, CodiEstatHTTP c)
        : ErrorTFC(w), codi_(c)
    {
        // buit
    }
    virtual ~ErrorHTTP() throw () {
        // buit
    }
    //! Obté el codi d'error HTTP
    CodiEstatHTTP codi() const { return codi_; }
private:
    //! El codi de l'error HTTP
    CodiEstatHTTP codi_;
};

/*!
 * \brief Petició HTTP incorrecta (familia d'errors <tt>4xx</tt>)
 */
class ErrorHTTPBadRequest : public ErrorHTTP {
public:
    explicit ErrorHTTPBadRequest(const std::string & w = "Petició HTTP incorrecta",
                                 CodiEstatHTTP c = CODI_400)
        : ErrorHTTP(w, c)
    {
        // buit
    }
};

/*!
 * \brief Error intern del servidor (familia d'errors <tt>5xx</tt>)
 */
class ErrorHTTPIntern: public ErrorHTTP {
public:
    explicit ErrorHTTPIntern(const std::string & w = "Error intern del servidor",
                             CodiEstatHTTP c = CODI_500)
        : ErrorHTTP(w, c)
    {
        // buit
    }
};

// ---------------------- Ajudes per al desenvolupament ---------------------- //

//! \brief Expressió evaluable a string que es resol a "nomfitxer:linia"
#define FITXER_I_LINIA_ \
                (std::string(__FILE__":")+boost::lexical_cast<std::string>(__LINE__))

/*!
 * \brief Llençat per mètodes no implementats
 */
class ErrorNoImplementat : public ErrorTFC {
public:
    explicit ErrorNoImplementat(const std::string & w = "No implementat")
        : ErrorTFC(w)
    {
        // buit
    }
};

// FIXME: Definició ràpida d'excepcions durant el desenvolupament
#define DEFINEIX_EXCEPCIO(nom, base) struct nom : public base {\
										explicit nom(const std::string &w = "") : base(w) {} \
									 }

} // ns tfc

#endif // _EXCEPCIONS_H_

// vim:set ts=4 et ai: //
