#if !defined(_TFCWEB_LOG_H_)
#define _TFCWEB_LOG_H_

/*!
 * \file
 * \author Toni Corvera
 * $Id: Log.h 483 2013-06-17 17:41:04Z toni $
 *
 * <h3>Logging</h3>
 * El mecanisme de \em logging és molt més simple d'utilitzar, del que la
 * interfície pot donar a entrendre.
 * Està dissenyat entorn \em missatges o \em entrades del log.
 *
 * Un objecte \ref tfc::Log únic controla els destins a on s'enviaràn els missatges
 * de log.
 * L'accés a l'objecte Log es fa mitjançant \ref tfc::Log::singleton(), i s'en
 * poden associar destins tals com fitxers o \em streams mitjançant \ref tfc::Log::associa()
 *
 * Els missatges individuals es creen amb \ref log() i
 * s'hi poden inserir valors com si es tractés d'un flux estàndard:
 * \code log() << "Missatge de log " << algun_objecte; \endcode
 *
 * Un missatge de log es \em finalitza inserint-hi \ref tfc::commit():
 * \code log() << "Missatge finalitzat" << commit; \endcode
 * En inserir \ref tfc::commit() el missatge s'envia automàticament a tots els destins
 * associats a l'objecte \ref tfc::Log.
 *
 * Cada destí i cada missatge tenen un nivell (\ref tfc::NivellLog), els missatges només
 * s'enviaran als destins que tinguin un nivell igual o inferior al seu.
 *
 * \todo TODO: Els manipuladors parametritzats ja no són necessaris perquè es poden inserir directament els nivells
 * \internal
 * L'ús de manipuladors de \ref tfc::MissatgeLog està inspirat en els manipuladors
 * de sortida de la biblioteca estàndard (p.e. \c std::endl): quan s'insereixen
 * en el \ref tfc::MissatgeLog el modifiquen o provoquen algun efecte com a conseqüència.
 *
 * Com en la biblioteca estàndard hi ha dos tipus de manipuladors, els simples
 * (com \c std::endl) i els parametritzats (com \c std::setfill)
 *
 * \sa Ús de manipuladors i generadors de manipuladors:
 *     http://stdcxx.apache.org/doc/stdlibug/33-3.html
 */
/*
 * Copyright (c) 2012 Toni Corvera
 *
 * This file is part of TFCWeb.
 *
 * TFCWeb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TFCWeb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TFCWeb. If not, see <http://www.gnu.org/licenses/>.
 */
// La construcció de missatges combinada amb una impressió "atòmica" hauria de permetre
// treballar cocurrentment sense problema

#include "Excepcions.h"

#include <fstream>
#include <memory>
#include <ostream>
#include <sstream>
#include <stdexcept>
#include <string>
#include <vector>
#include <boost/filesystem.hpp>
#include <boost/thread.hpp>

namespace tfc {

///////////////////////////////////////////////////////////////
// EXCEPCIONS
///////////////////////////////////////////////////////////////

/*!
 * \brief Indica que s'ha intentat utilitzar com a fitxer no vàlid
 *        (directori o no escribible)
 */
class ErrorLogNoValid : public ErrorTFC {
public:
    explicit ErrorLogNoValid(const std::string & w = "Fitxer de log no vàlid")
        : ErrorTFC(w)
    {
        // buit
    }
};

/*!
 * \brief Indica que s'ha intentat operar sobre un destí de logging no associat
 *        a l'objecte \c Log
 */
class ErrorDestiNoAssociat : public ErrorTFC {
public:
    explicit ErrorDestiNoAssociat(const std::string & w = "Destí de logging desconegut")
        : ErrorTFC(w)
    {
        // buit
    }
};

///////////////////////////////////////////////////////////////
// MISSATGES DE LOGGING
///////////////////////////////////////////////////////////////

/*!
 * \brief Nivells de logging
 * \sa nivell(NivellLog)
 */
enum NivellLog {
    NivellDebug = 1,
    NivellInfo,
    NivellWarning,
    NivellError,
    NIVELL_PER_DEFECTE_LOG = NivellDebug, // FIXME
    NIVELL_MES_SILENCIOS = NivellError,
    NIVELL_MES_VERBOS = NivellDebug,
    NIVELL_PER_DEFECTE_MISSATGE = NivellWarning
};

// Forward declarations
class Log;
class MissatgeLog;
/*!
 * \brief Interfície dels manipuladors parametritzats
 * \sa nivell()
 */
class ManipuladorParametritzat;
/*!
 * \brief Utilitzat internament en la implementació de \ref Log
 */
class DestiLogging;

/*!
 * \brief Interfície de creació de missatges de log
 *
 * Ús: \code log() << "Missatge " << ... << commit; \endcode
 */
MissatgeLog log();

/*!
 * \brief Manipulador final de missatge log: Imprimir la línia actual
 *
 * Quan s'insereix \c commit a \p ml es termina la línia i es força
 * la sincronització.\n
 * Exemple:
 * \code log() << "Hola" << commit; \endcode
 */
void commit(MissatgeLog & ml);

/*!
 * \brief Manipulador de missatge de log: Cancel·lar la línia actual de \p ml
 */
MissatgeLog& rollback(MissatgeLog & ml);

/*!
 * \brief Manipulador parametritzat: Indicació del nivell de \em logging de
 *        l'entrada
 * \remarks Inserible en un missatge de logging
 * \sa NivellLog
 *
 * Exemple: \code
 * log() << nivell(NIVELL_ERROR) << "Això és un error" << commit;
 * \endcode
 */
std::shared_ptr<ManipuladorParametritzat> nivell(NivellLog n);

/*!
 * \brief Missatge individual de log
 * \remarks Els missatges es creen mitjançant \ref log()
 */
class MissatgeLog {
public:
    /*!
     * \brief Operador d'inserció.
     */
    template<typename T> MissatgeLog& operator<<(const T & t) {
        oss_ << t;
        return *this;
    }

    /*!
     * \brief Canvi directe de nivells
     */
    MissatgeLog& operator<<(NivellLog n) {
        nivell(n);
        return *this;
    }

    /*!
     * \brief Operador d'inserció sobrecarregat per a manipuladors de MissatgeLog.
     * \param manipulador Manipulador a aplicar
     * \sa rollback
     * \remarks <tt>(MissatgeLog&(*)(MissatgeLog&)</tt> és la signatura dels
     *          manipuladors de MissatgeLog
     */
    MissatgeLog& operator<<(MissatgeLog&(*manipulador)(MissatgeLog&)) {
        return manipulador(*this);
    }

    /*!
     * \brief Operador d'inserció de manipuladors parametritzats
     * \remarks Permet inserir de manera natural aquest tipus de manipuladors
     */
    MissatgeLog& operator<<(const std::shared_ptr<ManipuladorParametritzat> &);

    /*!
     * \copydoc Operador d'inserció per a manipuladors \em finals de MissatgeLog.
     * \param finalitzador Manipulador a aplicar
     * \sa commit
     * \sa operator<<(MissatgeLog&(*)(MissatgeLog&))
     */
    void operator<<(void(*finalitzador)(MissatgeLog&)) {
        finalitzador(*this);
    }

    /*!
     * \brief Operador d'inserció sobrecarregat per a manipuladors estàndard.
     * \sa operator<<()
     * \remarks <tt>(std::ostream&(*)(std::ostream&)</tt> és la signatura dels
     *          manipuladors de sortida
     */
    MissatgeLog& operator<<(std::ostream&(*tt)(std::ostream&)) {
        oss_ << tt;
        return *this;
    }

    //! \brief Text del missatge
    std::string str() const { return oss_.str(); }

    //! \brief Obté el nivell de logging del missatge
    NivellLog nivell() const { return nivell_; }

    //! \brief Modifica el nivell de logging del missatge
    void nivell(NivellLog n) { nivell_ = n; }
private:
    MissatgeLog()
        : nivell_(NIVELL_PER_DEFECTE_MISSATGE), oss_()
    {
        // buit
    }
    MissatgeLog(const MissatgeLog & m)
        : nivell_(m.nivell_), oss_(m.oss_.str())
    {
        // buit
    }

    // Construcció
    friend MissatgeLog log();
    // Detalls d'implementació en forma de manipuladors:
    friend void commit(MissatgeLog &ml);
    friend MissatgeLog& rollback(MissatgeLog & ml);

    NivellLog nivell_; //!< Nivell del missatge
    std::ostringstream oss_; //!< Buffer
};

/*!
 * \brief Interfície de control de Logging.
 * \remarks Aquesta classe no es pot instanciar directament, fa servir el
 *          patró "Singleton"
 * \remarks \c Log pot tenir associats diferents destins, per defecte no en
 *          tè cap.
 *
 * Exemple: \code
 * Log & l = Log::singleton();
 * l.associa(std::cerr);
 * log() << "Linia de logging #" << 1 << commit; // Envia a Log::singleton()
 * \endcode
 * Produirà: \code "[MARCA DE TEMPS] Linia de logging #1" \endcode
 * \todo TODO: Implementar categories
 */
class Log {
public:
    /*!
     * \brief Interfície cap a la única instància
     */
    static Log& singleton();

    /*!
     * \brief Afegeix \p os com a destí de logging
     * \note Es poden definir múltiples destins de logging
     */
    void associa(std::ostream & os, NivellLog n = NIVELL_PER_DEFECTE_LOG);

    /*!
     * \brief Afegeix el fitxer \p f com a destí de logging
     * \remarks El fitxer s'obrirà en mode d'annexió (\em append)
     * \pre !boost::filesystem::exists(p) || boost::filesystem::is_regular_file(p)
     * \throws ErrorLogNoValid Si no es pot obrir \p f o no és un fitxer vàlid
     */
    void associa(const boost::filesystem::path & f, NivellLog n = NIVELL_PER_DEFECTE_LOG)
                                                                       throw (ErrorLogNoValid);

    /*!
     * \brief Elimina \p os de la llista de destins
     * \remarks Desassociar un destí que no s'ha associat no supossa un error
     * \todo TODO: No implementat!
     */
    void desassocia(std::ostream & os);

    /*!
     * \brief Impressió d'un missatge.
     * \remarks Cada missatge inclourà una marca de temps
     */
    void imprimeix(const MissatgeLog & ml);

    /*!
     * \brief Obté el nivell actual de sortida de \em logging associat a \p d
     *
     * Els missatges marcats amb un nivell inferior s'ignoraran
     * \throws ErrorDestiNoAssociat si \p d no és un destí associat a aquest objecte
     */
    NivellLog nivell_sortida(const std::ostream & d) const
                                                        throw (ErrorDestiNoAssociat);
    //! \copydoc nivell_sortida(const std::ostream &) const
    NivellLog nivell_sortida(const boost::filesystem::path & d) const
                                                        throw (ErrorDestiNoAssociat);
    /*!
     * \brief Canvia el nivell actual de sortida de \em logging del destí \p d.
     *
     * Els missatges marcats amb un nivell inferior s'ignoraran
     * \throws ErrorDestiNoAssociat si \p d no és un destí associat a aquest objecte
     */
    void nivell_sortida(const std::ostream & d, NivellLog nn)
                                                        throw (ErrorDestiNoAssociat);
    //! \copydoc nivell(const std::ostream&, NivellLog)
    void nivell_sortida(const boost::filesystem::path & d, NivellLog nn)
                                                        throw (ErrorDestiNoAssociat);

    /*!
     * \brief Canvia el nivell actual de sortida de \em logging de tots els destins
     */
    void nivell_sortida(NivellLog nn);

protected:
    Log()
        : destins_()
    {
        // buit
    }

    //! Marca de text formatada
    static std::string marca_de_temps();
private:
    //! \brief Constructor de còpia, \b esborrat
    Log(const Log & l) { throw ErrorNoImplementat(FITXER_I_LINIA_); }

    //! \brief Operador d'assignació, \b esborrat
    Log& operator=(const Log &) {
        throw ErrorNoImplementat(FITXER_I_LINIA_);
        return *this;
    }

    std::vector< std::shared_ptr<DestiLogging> > destins_; //!< Destins associats al Log
};

} // ns tfc

#endif // _TFCWEB_LOG_H_

// vim:set ts=4 et ai: //
