#if !defined(_MISSATGE_HTTP_H_)
#define _MISSATGE_HTTP_H_

/*!
 * \file
 * \author Toni Corvera
 *
 * $Id: MissatgeHTTP.h 483 2013-06-17 17:41:04Z toni $
 */
/*
 * Copyright (c) 2012 Toni Corvera
 *
 * This file is part of TFCWeb.
 *
 * TFCWeb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TFCWeb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TFCWeb. If not, see <http://www.gnu.org/licenses/>.
 */

#include "CapsaleraHTTP.h"
#include "portabilitat.h"

#include <memory>
#include <vector>

namespace tfc {

// FIXME: RFC 2068 també?
// L'RFC recomana (no exigeix) enviar les capçaleres amb un ordre determinat
// (de més generals a més concretes respecte el contingut del missatge)
//
// Interessa conservar l'ordre d'inserció, si es fa servir una taula hash
// caldrà algún mètode de fer-ho:
// - afegir un vector<string> amb l'ordre d'inserció
// - utilitzar boost::multi_index
//   <http://www.boost.org/doc/libs/1_49_0/libs/multi_index/doc/index.html>
//   <http://stackoverflow.com/questions/1098175/a-stdmap-that-keep-track-of-the-order-of-insertion>

/*!
 * \brief Classe base de missatges HTTP
 */
template<typename TCos>
class MissatgeHTTP {
public:
    typedef std::vector<CapsaleraHTTP> tipus_llista_capsaleres;

    /*!
     * \brief Obté el valor d'una capçalera
     * \param nom Nom de la capçalera
     * \param[out] definida (opcional) \c true si la capçalera es troba, \c false altrament
     * \remarks \p nom no distingeix majúscules de minúscules
     * \returns Capçalera \p nom, cadena buida si no està definida
     */
    std::string capsalera(const std::string & nom, bool * definida = 0) const;

    //! Operador d'igualtat
    virtual bool operator==(const MissatgeHTTP<TCos> & m) const {
        if (this == &m) {
            return true;
        }
        return capsaleres_ == m.capsaleres_; // FIXME: == ó std::equal() ? Test case
    }

    /*!
     * Afegeix una nova capçalera al missatge
     * \remarks \p nom no distingeix majúscules de minúscules
     */
    void afegeix_capsalera(const std::string & nom, const std::string & valor);

    /*!
     * \brief Afegeix múltiples capçaleres al missatge
     * \remarks La seqüència passada ha de fer referència a una col·lecció de
     *          \c std::pair<std::string, std::string> amb {nom, valor}
     */
    template<typename Iterador>
    void afegeix_capsaleres(const Iterador & inici, const Iterador & final);

    /*!
     * Obté la llista de capçaleres incloses en el missatge
     * \sa afegeix_capsalera()
     * \returns La llista
     * \todo XXX: Defensiu?: Referencia no modificable
     * \deprecated Reemplaçar per Iterator& capsaleres() ?
     */
    const tipus_llista_capsaleres& capsaleres() const {
        return capsaleres_;
    }

    /*!
     * \brief Defineix el cos del missatge
     * \param cos Nou cos del missatge
     */
    void defineix_cos(std::shared_ptr<TCos> cos) {
        cos_ = cos;
    }

    /*!
     * \brief Imprimeix les capçaleres afegides al missatge.
     * Imprimeix les capçaleres, una per linia, en el format "Nom: Valor"
     * \param os Stream on imprimir
     * \param le Cadena a insertar al final de cada línia
     */
    void imprimeix_capsaleres(std::ostream &os, const std::string & le = "\n") const;

    /*!
     * \brief Obté el cos
     * \returns Cos
     */
    virtual std::weak_ptr<TCos> cos() const { return cos_; }

    bool te_cos() const { return 0 != cos_.get(); }

    /*!
     * \brief Elimina una capçalera del missatge, si existeix
     * \return \c true si existía, \c false si no
     */
    bool elimina_capsalera(const std::string & nom) throw();

protected:
    MissatgeHTTP()
        : cos_()
    {
        // buit
    }

    /*!
     * \brief Constructor de còpia
     * \todo FIXME: Inserir còpies? \c std::copy insereix còpies?
     */
    MissatgeHTTP(const MissatgeHTTP & m);
private:

    /*!
     * \brief "Functor" per imprimir cada element d'una col·lecció passant-lo
     *        a \c std::for_each.
     * \sa MissatgeHTTP::imprimeix_capsaleres()
     */
    struct Impressor;

    /*!
     * \brief Capçaleres afegides al missatge.
     * \todo XXX: De moment no es distingeix entre capçaleres generals i
     *            específiques. Important?
     */
    tipus_llista_capsaleres capsaleres_;

#if 0
    // L'RFC no obliga però recomana ordenar les capçaleres (de més a menys generals)
    // amb aquesta intenció interessa mantenir l'ordre d'inserció. Per a fer-ho
    // cap utilitzar un contenidor addicional (ni map ni unordered_map mantenen l'ordre
    // d'inserció), o un dels contenidors multi-clau de boost
    /*!
     * \brief Capçaleres definides al missatge
     * \details
     * La clau és el nom de la capçalera en minúscules mentre que el valor
     * conserva la forma utilitzada en la primera inserció
     */
    std::unordered_map<std::string, CapsaleraHTTP> capsaleres_;
    /*!
     * \brief Ordre d'inserció de les capçaleres.
     * \details
     * Els elements es corresponen amb les claus definides a \ref capsaleres_
     */
    std::vector<std::string> ordre_capsaleres_;
#endif

    std::shared_ptr<TCos> cos_; //!< Cos del missatge
};

} // ns tfc

// Implementació (inclosa directament per ser una classe template)
#include "MissatgeHTTP.hxx"

#endif // _MISSATGE_HTTP_H_

// vim:set ts=4 et ai: //
