
/*!
 * \file
 * \author Toni Corvera
 *
 * $Id: Permisos.cc 483 2013-06-17 17:41:04Z toni $
 */
/*
 * Copyright (c) 2012 Toni Corvera
 *
 * This file is part of TFCWeb.
 *
 * TFCWeb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TFCWeb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TFCWeb. If not, see <http://www.gnu.org/licenses/>.
 */

#include "Permisos.h"

using namespace boost::filesystem;

namespace {

enum {
    PERMIS_LIMIT = EXECUCIO|LECTURA|ESCRIPTURA|PROTEGIT, //!< Utilitzat en invariants

    MASCARA_PERMIS = PERMIS_LIMIT|PRIVAT //!< Màscara dels bits significatius per FlagsPermis
};

} // ns anònim


namespace tfc {

void Cerber::defineix_regla(const ruta_t & ruta, FlagsPermis perm) {
    perm &= MASCARA_PERMIS; // Desactiva els bits irrellevants
    // Rebre 0 es considera equivalent a PRIVAT
    if (0 == perm || 0 != (PRIVAT & perm)) {
        // Privat: La resta de permisos no tenen sentit
        perm = PRIVAT;
    }
    // INVARIANTS:
    assert( 0 != perm );
    assert( perm <= PERMIS_LIMIT );

    // Si ja està definida, les combinem (PRIVAT no combina amb res)
    if (PRIVAT != perm) {
        auto it = regles_.find(ruta);
        if (regles_.end() != it) {
            perm = perm | it->second;
        }
    }

    // regles_.insert( make_pair(ruta, perm) ); // No sobre-escriu!
    regles_[ruta] = perm;
}

CodiEstatHTTP Cerber::acces(const path & ruta, TipusPeticioHTTP t) const {
    const FlagsPermis p = permis(ruta);
    assert( 0 != p ); // INVARIANT
    assert( p <= PERMIS_LIMIT ); // INVARIANT
    if (0 != (p & PRIVAT)) {
        return CODI_403;
    }
    if (0 != (p & PROTEGIT)) {
        return CODI_401;
    }
    Permis mascara;
    switch (t) {
    case GET:
    case HEAD:
    case POST:
    case OPTIONS:
    case TRACE:
    case CONNECT:
        mascara = LECTURA;
        break;
    case PUT:
    case METODE_DELETE:
        mascara = ESCRIPTURA;
        break;
    default:
        assert( !"Error intern" );
        return CODI_403;
    }

    if (0 == (p & mascara)) {
        return CODI_405;
    }

    return CODI_200;
}

FlagsPermis Cerber::permis(const path & ruta) const {
    const path buit = path();
    path r(ruta);
    do { // Comprovem la ruta inicial i les superiors
        if (exists(r) &&!utils::es_llegible(r)) {
            // Si existeix però els permisos d'accés no permeten l'accés,
            // és com si estigués prohibit l'accés
            return PRIVAT;
        }
        auto it = regles_.find(r);
        if (regles_.end() != it) {
            // Trobat!
            return it->second;
        }
        r = r.parent_path();
    } while (buit != r);
    return per_defecte_ ;
}

bool Cerber::executable(const ruta_t & ruta) const {
    const FlagsPermis p = permis(ruta);
    assert( 0 != p ); // INVARIANT
    assert( p <= PERMIS_LIMIT ); // INVARIANT
    return 0 != ( (LECTURA|EXECUCIO) & p );
}

// friend Cerber
ostream& operator<<(ostream & os, const Cerber & c) {
    os << "Permisos {"
       << "\n\tdefecte = " << nom_permis(c.per_defecte_);
    for (auto it=c.regles_.begin(); it != c.regles_.end(); ++it) {
        os << "\n\t" << it->first << " = " << nom_permis(it->second) ;
    }
    return os << "\n}";
}

string nom_permis(FlagsPermis f) {
    f &= MASCARA_PERMIS; // Desactiva els bits irrellevants
    if (PRIVAT == f) {
        return "Privat";
    }
    string n;
    if (0 != (f & PROTEGIT)) {
        n += ",Protegit";
    }
    if (0 != (f & LECTURA)) {
        n += ",Lectura";
    }
    if (0 != (f & ESCRIPTURA)) {
        n += ",Escriptura";
    }
    if (0 != (f & EXECUCIO)) {
        n += ",Execució";
    }
    if (n.empty()) {
        return n;
    }
    assert( n.length() > 0 );
    n.erase(0, 1);
    return n;
}

} // ns tfc

// vim:set ts=4 et ai: //
