#if !defined(_PERMISOS_H_)
#define _PERMISOS_H_

/*!
 * \file
 * \author Toni Corvera
 *
 * $Id: Permisos.h 483 2013-06-17 17:41:04Z toni $
 */
/*
 * Copyright (c) 2012 Toni Corvera
 *
 * This file is part of TFCWeb.
 *
 * TFCWeb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TFCWeb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TFCWeb. If not, see <http://www.gnu.org/licenses/>.
 */

#include "PeticioHTTP.h"

#include <iostream>
#include <boost/filesystem.hpp>
#include <boost/unordered_map.hpp>

namespace tfc {

//! Modelació dels possibles permisos
enum Permis {
    PRIVAT = 0x01,     //!< No es permet cap mena d'accés
    PROTEGIT = 0x02,   //!< Només amb contrasenya
    LECTURA = 0x04,    //!< Permís de lectura (GET, HEAD, POST...)
    ESCRIPTURA = 0x08, //!< Permís d'escriptura (PUT, DELETE)
    EXECUCIO = 0x10    //!< Permís d'execució de CGIs
};
//! Composició de valors a partir de Permis (p.e. \c PROTEGIT|LECTURA)
typedef unsigned short FlagsPermis;

/*!
 * \brief Forma textual dels permisos
 */
std::string nom_permis(FlagsPermis f);

class Permisos;

class Cerber {
public:
    typedef boost::filesystem::path ruta_t;

    /*!
     * \attention Si es fa servir un permís lax per defecte, és important
     *            definir una regla "limit" que prohibeixi l'accés per sobre
     *            de l'arrel del servidor, exemple:
     *            \code
     *            Configuracio cfg;
     *            Cerber cerber;
     *            cerber.defineix_regla(cfg.arrel().parent_path(), PRIVAT);
     *            \endcode
     */
    explicit Cerber(FlagsPermis permis_per_defecte = LECTURA)
        : per_defecte_(permis_per_defecte)
    {
        // buit
    }

    Cerber(const Cerber & c)
        : regles_(c.regles_),
          per_defecte_(c.per_defecte_)
    {
        // buit
    }

    Cerber& operator=(const Cerber & c) {
        regles_ = c.regles_;
        per_defecte_ = c.per_defecte_;
        return *this;
    }

    /*!
     * \brief Defineix una regla d'accés
     * \remarks Successives crides es combinen, excepte si \p perms conté
     * \c PRIVAT p.e.
     * \code
     * c.defineix_regla(r, LECTURA);
     * c.defineix_regla(r, ESCRIPTURA);
     * c.permis() == LECTURA|ESCRIPTURA
     * c.defineix_regla(r, PRIVAT);
     * c.permis() == PRIVAT
     * \endcode
     */
    void defineix_regla(const ruta_t & ruta, FlagsPermis perms);

    /*!
     * \brief Comprova quins permisos té la ruta \p ruta
     * \return \c CODI_200 si es permet l'accés,
     *         \c CODI_401 si l'accés és permés però requereix contrasenya,
     *         \c CODI_403 si l'accés és prohbit,
     *         \c CODI_405 si el mètode \p metode no es permet per aquesta ruta
     * \remarks La comprovació de permís es pot simplificar comparant amb CODI_400:
     * \code permis(r,m) < CODI_400 // -> OK
     *       permis(r,m) >= CODI_400 // -> No permés
     * \endcode
     * \sa permis()
     */
    CodiEstatHTTP acces(const ruta_t & ruta, TipusPeticioHTTP t) const;

    /*!
     * Com \ref acces() però obtenint un permís simbòlic en comptes d'un codi d'estat
     * \sa acces()
     */
    FlagsPermis permis(const ruta_t & ruta) const;

    /*!
     * \brief Comprova específicament si un fitxer té permís d'execució I de lectura
     */
    bool executable(const ruta_t & ruta) const;
private:
    boost::unordered_map<ruta_t, FlagsPermis> regles_;
    FlagsPermis per_defecte_;

    // Impressió de les regles
    friend std::ostream& operator<<(std::ostream&, const Cerber&);
};

}

#endif // _PERMISOS_H_

// vim:set ts=4 et ai: //
