#if !defined(_PETICIO_HTTP_H_)
#define _PETICIO_HTTP_H_

/*!
 * \file
 * \author Toni Corvera
 *
 * $Id: PeticioHTTP.h 483 2013-06-17 17:41:04Z toni $
 */
/*
 * Copyright (c) 2012 Toni Corvera
 *
 * This file is part of TFCWeb.
 *
 * TFCWeb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TFCWeb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TFCWeb. If not, see <http://www.gnu.org/licenses/>.
 */

#include "CapsaleraHTTP.h"
#include "CosPeticio.h"
#include "Excepcions.h"
#include "MissatgeHTTP.h"
#include "portabilitat.h"
#include "Uri.h"

#include <algorithm>
#include <cassert>
#include <list>
#include <map>
#include <ostream>

namespace tfc {

/*!
 * Tipus de peticions HTTP/1.1
 * \sa NomTipusPeticioHTTP::nom_tipus()
 */
enum TipusPeticioHTTP {
    GET = 1,
    POST,
    HEAD,
    OPTIONS, // XXX: CONNECT ??
    TRACE,
    PUT,
    METODE_DELETE, // DELETE és una macro a Windows! (winnt.h)
    CONNECT, // Introduït en l'RFC 2616
    //! Tipus no definit o no suportat
    desconegut
};

/*!
 * \brief Indica que s'ha trobat un mètode HTTP desconegut en una petició
 */
class ErrorMetodeDesconegut : public ErrorTFC {
public:
    explicit ErrorMetodeDesconegut(const std::string & w = "Mètode HTTP desconegut")
        : ErrorTFC(w)
    {
        // buit
    }
};


/*!
 * \brief Encapsulament de la traducció de tipus de petició al seu nom
 */
namespace NomTipusPeticioHTTP {
    /*!
     * \brief Obté el nom (verb) d'un tipus de petició
     * \param tipus Tipus simbòlic de petició
     */
    const std::string& nom(TipusPeticioHTTP tipus);

    /*!
     * \brief Obté el tipus corresponent a un nom (verb)
     * \param nom Nom del tipus de petició
     * \post tipus(X) != TipusPeticioHTTP::desconegut
     */
    TipusPeticioHTTP tipus(const std::string & nom) throw (ErrorMetodeDesconegut);
}

/*!
 * \brief Indicació de versió d'HTTP
 */
enum VersioHTTP {
    HTTP_DESCONEGUT,
    HTTP_0_9,
    HTTP_1_0,
    HTTP_1_1,
    HTTP_1_X, // HTTP 1.x > 1.1
    HTTP_SUPERIOR // HTTP >= 2.0
};
//! Impressió transparent dels valors simbòlics de VersioHTTP
std::ostream& operator<<(std::ostream & os, tfc::VersioHTTP v);

/*!
 * \brief Representació d'una petició HTTP
 */
class PeticioHTTP : public MissatgeHTTP<CosPeticio> {
public:
    PeticioHTTP(TipusPeticioHTTP tipus,
                const Uri & ruta,
                VersioHTTP versio)
        : MissatgeHTTP<CosPeticio>(), tipus_(tipus), ruta_(ruta), versio_(versio)
    {
		// buit
    }

    //! Constructor de còpia
    PeticioHTTP(const PeticioHTTP& p)
        : MissatgeHTTP<CosPeticio>(p), tipus_(p.tipus_), ruta_(p.ruta_), versio_(p.versio_)
    {
		// buit
    }

    // Getters

    //! Tipus de la petició
    TipusPeticioHTTP tipus() const { return tipus_; }

    //! Versió HTTP
    VersioHTTP versio() const { return versio_; }

    //! Ruta de la petició
    const Uri& ruta() const { return ruta_; }

    // Setters

    //! Defineix el tipus de la petició
    void tipus(TipusPeticioHTTP t) { tipus_ = t; }

    //! Defineix la versió HTTP
    void versio(VersioHTTP versio) { versio_ = versio; }

    //! Defineix la ruta sol·licitada
    void ruta(const Uri & ruta) { ruta_ = ruta; }
private:
    TipusPeticioHTTP tipus_;
    Uri ruta_;
    VersioHTTP versio_;

    /*!
     * \brief Representació textual
     * \todo XXX: No necessari? Al menys útil durant el desenvolupament
     */
    friend std::ostream& operator<< (std::ostream & os, const PeticioHTTP & r) {
        os << NomTipusPeticioHTTP::nom(r.tipus()) << ", " << r.ruta()
           << ", " << r.versio() << " [\n";
        r.imprimeix_capsaleres(os);
        return os << "]\n";
    }
};

} // ns tfc

#endif // _PETICIO_HTTP_H_

// vim:set ts=4 et ai: //
