#if !defined(_RECURS_H_)
#define _RECURS_H_

/*!
 * \file
 * \author Toni Corvera
 *
 * $Id: Recurs.h 483 2013-06-17 17:41:04Z toni $
 */
/*
 * Copyright (c) 2012 Toni Corvera
 *
 * This file is part of TFCWeb.
 *
 * TFCWeb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TFCWeb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TFCWeb. If not, see <http://www.gnu.org/licenses/>.
 */

#include "CosResposta.h"
#include "Excepcions.h"
#include "portabilitat.h"
#include "Transformacions.h"
#include "Uri.h"

#include <ostream>
#include <stdexcept>
#include <string>
#include <boost/filesystem.hpp>

namespace tfc {

/*!
 * \brief Indica que s'ha produït un error en intentar llegir els continguts
 *        d'un fitxer
 */
class ErrorLlegintFitxer : public ErrorTFC {
public:
    ErrorLlegintFitxer(const std::string & w = "Error llegint fitxer")
        : ErrorTFC(w)
    {
        // buit
    }
};

/*!
 * \brief Representació d'un recurs (fitxer o directori)
 * Abans d'instanciar cal inicialitzar estàticament amb \sa Recurs::init()
 */
class Recurs : public CosResposta {
public:
    /*!
     * \param arrel_servidor Directori arrel del servidor
     * \param peticio Petició enviada pel client
     * \throws ErrorRutaInexistent si ruta_peticio no es correspon a una ruta
     *                             existent en el servidor
     * \pre \p arrel_servidor existeix, és un directori, és absoluta i canònica
     * \post \a Recurs.fisica() existeix en el disc, és absoluta i canònica
     * \sa Recurs::init()
     */
    explicit Recurs(const PeticioHTTP & peticio,
                    const boost::filesystem::path & arrel_servidor) throw (ErrorRutaInexistent);
    //! \brief Constructor de còpia
    Recurs(const Recurs & r) throw ()
        : CosResposta(r),
          arrel_servidor_(r.arrel_servidor_),
          fisica_(r.fisica_)
    {
        using namespace boost::filesystem;
        // INICIALITZACIÓ CORRECTA
        assert( path() != arrel_servidor_ );
        // INVARIANT
        assert( canonical(r.fisica_) == r.fisica_ );
    }
    //! \brief Operador d'assignació
    Recurs& operator=(const Recurs & r) throw () {
        CosResposta::operator=(r);
        using namespace boost::filesystem;
        // INICIALITZACIÓ CORRECTA
        assert( path() != arrel_servidor_ );
        // INVARIANT
        assert( canonical(r.fisica_) == r.fisica_ );

        arrel_servidor_ = r.arrel_servidor_;
        fisica_ = r.fisica_;
        return *this;
    }

    //! \brief Ruta física del recurs en el servidor
    const boost::filesystem::path& fisica() const { return fisica_; }
    //! \brief Ruta de la petició normalitzada respecte l'arrel del servidor
    const Uri publica() const { throw ErrorNoImplementat(); return Uri(""); }

    /*!
     * \copydoc CosResposta::envia(boost::asio::ip::tcp::socket&)
     * \throws ErrorLlegintFitxer Si es produeix un error en la lectura del recurs
     */
    virtual void envia(boost::asio::ip::tcp::socket & socket)
                                const throw (std::runtime_error);

    virtual tipus_capsaleres capsaleres() const;

protected:
    using CosResposta::envia;
private:
    /*!
     * Genera la etiqueta d'entitat (ETag)
     */
    std::string entity_tag() const;

    /*!
     * Impressió.
     * \deprecated ? Només útil durant desenvolupament
     */
    friend std::ostream& operator<<(std::ostream& os, const Recurs & r) {
        return os << "Recurs ( Petició: " << r.peticio()
                  << ", Disc: " << r.fisica()
                  //<< ", En servidor: " << std::string(r.publica())
                  << " )";
    }

    boost::filesystem::path arrel_servidor_;
    boost::filesystem::path fisica_;
};

} // ns tfc

#endif // _RECURS_H_

// vim:set ts=4 et ai: //
