#if !defined(_TFCWEB_THREADPOOL_H_)
#define _TFCWEB_THREADPOOL_H_

/*!
 * \file
 * \author Toni Corvera
 *
 * $Id: ThreadPool.h 483 2013-06-17 17:41:04Z toni $
 */
/*
 * Copyright (c) 2012 Toni Corvera
 *
 * This file is part of TFCWeb.
 *
 * TFCWeb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TFCWeb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TFCWeb. If not, see <http://www.gnu.org/licenses/>.
 */

#include "CuaConcurrent.h"
#include "Excepcions.h"
#include "utils.h"

#include <cstddef>
#include <queue>
#include <vector>
#include <boost/any.hpp>
#include <boost/function.hpp>
#include <boost/thread.hpp>

namespace tfc {

class ErrorThreadPoolTancada : public ErrorTFC {
public:
    explicit ErrorThreadPoolTancada(const std::string & w = "Thread pool tancada")
        : ErrorTFC(w)
    {
        // buit
    }
};

/*!
 * \brief Implementació del patró \em ThreadPool
 *
 * En aquesta implementació un ThreadPool controla un nombre fix de fils
 * treballadors. Aquests processen tasques a mida que se n'afegeixen i romanen
 * a l'espera si no hi ha cap tasca. Si hi ha més tasques que fils, les últimes
 * en arrivar es quedaran en espera fins que algun dels fils treballadors
 * finalitzi la tasca que estigui executant.
 *
 * Les tasques han de ser cridables (p.e. funcions o objectes funció) que
 * no requereixin arguments.
 */
class ThreadPool {
public:
    // void fn() -> function<void (void)>
    typedef boost::function<void (void)> tipus_tasca; //!< Tipus de les tasques acceptades
    typedef CuaConcurrent<tipus_tasca> tipus_cua_tasques; //!< Tipus de la cua de tasques

    /*!
     * \brief Construeix el \em ThreadPool amb \p n_workers fils.
     * \param n_workers Nombre de fils a crear
     * \remarks Els fils s'inicien immediatament
     * \post Els \p n_workers fils s'han creat i estan executant-se a l'espera de
     *       tasques
     */
    explicit ThreadPool(const size_t n_workers);

    virtual ~ThreadPool();

    /*!
     * \brief Programa una tasca \p t
     * \throws ErrorThreadPoolTancada Si s'intenta programar una tasca un cop
     *         s'ha tancat l'objecte
     * \remarks Possibles tipus de \p t: funció, objecte funció o objecte funció
     *          obtingut amb \c std::mem_fn o \c boost::bind per exemple.
     * \sa tanca()
     * \sa termina()
     */
    void programa(tipus_tasca t) throw (ErrorThreadPoolTancada);

    //! \brief Es deixen d'acceptar tasques
    void tanca();

    //! \brief Comprova si s'ha tancat
    bool es_oberta() const { return !tancada_; }

    /*!
     * \brief Bloca el fil actual fins que terminin totes les tasques
     */
    void join();

    /*!
     * \brief Cancel·la les tasques pendents
     * \remarks Retorna immediatament
     */
    void termina();

    /*!
     * \brief Obté el nombre de tasques pendents
     */
    size_t num_tasques() const {
        return tasques_.size();
    }

private:
    void inicia(); //!< Inicia els fils

    const size_t n_threads_; //!< Nombre de fils
    std::vector< boost::thread * > threads_; //!< Fils
    tipus_cua_tasques tasques_; //!< Tasques programades \sa programa()
    bool tancada_; //!< Indica si s'accepten noves tasques
    volatile bool marca_terminacio_; //!< Indica als fils que han de finalitzar immediatament
};

} // ns tfc

#endif // _TFCWEB_THREADPOOL_H_

// vim:set ts=4 et ai: //
