#if !defined(_TRANSFORMACIONS_H_)
#define _TRANSFORMACIONS_H_

/*!
 * \file
 * \author Toni Corvera
 * Transformacions de text
 *
 * Per a permetre treballar sense fitxer temporals en què escriure les
 * dades transformades es defineixen les transformacions associades a
 * dos \em callbacks: un per extreure un bloc de dades de l'origen
 * (p.e. un fitxer) i un altre per escriure'l transformat a la
 * destinació (p.e. un socket).
 *
 * A més es defineixen uns callbacks genèrics per extreure d'un \c istream
 * i escriure en un \c ostream (p.e. es poden fer servir amb un fitxer
 * i \c cout per escriure'l per pantalla).
 *
 * $Id: Transformacions.h 483 2013-06-17 17:41:04Z toni $
 */
/*
 * Copyright (c) 2012 Toni Corvera
 *
 * This file is part of TFCWeb.
 *
 * TFCWeb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TFCWeb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TFCWeb. If not, see <http://www.gnu.org/licenses/>.
 */

#include "portabilitat.h" // zlib.h, ZLIB_CONST

#include <array>
#include <iostream>
#include <memory>
#include <string>
#include <vector>
#include <boost/function.hpp>

namespace tfc {

/*!
 * \brief Possibles estats en cridar Transformacio::destransforma()
 */
enum EstatTransformacioInversa {
    TRANSFORMAT, //!< Transformació correcta
    NO_TRANSFORMAT, //!< Transformació errònia (p.e. dades insuficients)
    FORMAT_INCORRECTE //!< Transformació errònia i no sol·lucionable amb més dades
};

/*!
 * \brief Interfície de les transformacions
 */
class Transformacio {
public:
    virtual ~Transformacio() {
        // buit
    }

    /*!
     * \brief Transforma un bloc de dades
     */
    virtual std::string transforma(const std::string &bloc) const = 0;

    /*!
     * \brief Genera un bloc final
     */
    virtual std::string finalitza() const = 0;

    /*!
     * \brief Inverteix la transformació
     * \remarks A diferencia de \ref transforma() està pensada per ser cridada
     *          sobre el text transformat al complet
     * \todo TODO: Fer servir \em callbacks
     */
    virtual std::string destransforma(const std::string & text_transformat,
                                      EstatTransformacioInversa & estat) const = 0;
};

/*!
 * \brief Classe base d'implementació de \ref AplicadorTransformacio,
 *        no instanciable
 * \sa AplicadorTransformacio
 * \internal Classe independent per permetre separar la declaració
 *           i la implementació de AplicadorTransformacio, què és un
 *           template
 */
class AplicadorTransformacioBase {
public:
    /*!
     * \brief Signatura del \em callback d'extracció.
     * Interfície:<br />
     * Sobreescriurà l'\c string passat amb el bloc de dades extret.
     * Retornarà \c true mentre s'hagi pogut extreure correctament.
     */
    typedef boost::function<bool (std::string&)> funcio_extractora;
    /*!
     * \brief Signatura del \em callback d'escriptura.
     * Interfície:<br />
     * Rebrà el bloc de dades extretes en la última operació a l'\c string passat.
     * Retornarà \c true mentre s'hagi pogut escriure correctament.
     */
    typedef boost::function<bool (const std::string&)> funcio_escriptora;

    //! \brief Funció extractora associada
    const funcio_extractora& extractora() const { return extractora_; }
    //! \brief Funció escriptora associada
    const funcio_escriptora& escriptora() const { return escriptora_; }

    /*!
     * \brief Transforma mentre sigui possible.
     * \remarks Transformarà blocs mentre la lectura i l'escriptura no donin
     *          error
     */
    void aplica();

    virtual ~AplicadorTransformacioBase();
protected:
    AplicadorTransformacioBase(const funcio_extractora & extractora,
                               const funcio_escriptora & escriptora)
        : extractora_(extractora), escriptora_(escriptora)
    {
        // buit
    }
    //! Transformació associada
    virtual Transformacio* transformacio() = 0;
private:
    const funcio_extractora extractora_;
    const funcio_escriptora escriptora_;
};

/*!
 * \brief Aplica una transformació a un origen i una destinació
 * \tparam TTransformacio Tipus de la transformacio.
 *         Ha de ser convertible en \ref Transformacio
 */
template<typename TTransformacio>
class AplicadorTransformacio : public AplicadorTransformacioBase {
public:
    AplicadorTransformacio(const funcio_extractora & extractora,
                           const funcio_escriptora & escriptora)
        : AplicadorTransformacioBase(extractora,escriptora), trans_()
    {
        // buit
    }
protected:
    //! \copydoc AplicadorTransformacioBase::transformacio()
    virtual Transformacio* transformacio() { return &trans_; }
private:
    TTransformacio trans_;
};

/*!
 * \brief Extracció de blocs des d'un \c istream
 */
class FuncioExtractoraIstream {
public:
    FuncioExtractoraIstream(std::istream & ist)
        : ist_(ist)
    {
        // buit
    }
    bool operator()(std::string &);
private:
    std::istream & ist_;
    enum {
        MIDA_BUFFER = 4096
    };
    std::array<char, MIDA_BUFFER> buffer_;
};

/*!
 * \brief Escriptura de blocs en un \c ostream
 */
class FuncioEscriptoraOstream {
public:
    FuncioEscriptoraOstream(std::ostream & ost)
        : ost_(ost)
    {
        // buit
    }
    bool operator()(const std::string &);
private:
    std::ostream & ost_;
};

/*!
 * \brief Transformació \em identity (nul·la)
 */
class TransformacioIdentity : public Transformacio {
public:
    virtual std::string transforma(const std::string & bloc) const { return bloc; }
    virtual std::string finalitza() const { return ""; }
    virtual std::string destransforma(const std::string & text,
                                      EstatTransformacioInversa & ok) const {
        ok = TRANSFORMAT;
        return text;
    }
};

/*!
 * \brief Transformació \em chunked
 * \sa http://tools.ietf.org/html/rfc2616#section-3.6.1
 */
class TransformacioChunked : public Transformacio {
public:
    virtual std::string transforma(const std::string & bloc) const;
    virtual std::string finalitza() const;
    virtual std::string destransforma(const std::string & text_transformat,
                                      EstatTransformacioInversa & estat) const;
};

class TransformacioDeflate : public Transformacio {
public:
    TransformacioDeflate()
        : inicialitzat_(CAP)
    {
        // buit
    }

    virtual std::string transforma(const std::string & bloc) const;
    virtual std::string finalitza() const;
    virtual std::string destransforma(const std::string & text_transformat,
                                      EstatTransformacioInversa & estat) const;
private:
    enum Mode {
        CAP,
        DEFLATE,
        INFLATE
    };
    // http://www.zlib.net/zlib_how.html:
    // "
    // Larger buffer sizes [respecte 16384] would be more efficient,
    // especially for inflate().
    // If the memory is available, buffers sizes on the order of 128K
    // or 256K bytes should be used.
    // "
    enum {
        MIDA_BUFFER_ZLIB = 256<<10
    };

    /*!
     * \brief Inicialitza una tasca de compressió o descompressió
     * \remarks Si hi ha una inicialitzada la desinicialitza
     * \param m Mode de treball, \c DEFLATE per a compressió,
     *          \c INFLATE per a compressió
     */
    bool inicialitza(Mode m) const;
    /*!
     * \brief Desinicialitza una tasca, si hi ha alguna inicialitzada
     */
    void desinicialitza() const;

    mutable z_stream strm_;
    mutable Mode inicialitzat_;
};

/*!
 * \brief Transformacio encadenada
 * Exemple: Transformació "deflate,chunked": \code
 * CadenaTransformacions cadena;
 * cadena.afegeix<TransformacioDeflate>();
 * cadena.afegeix<TransformacioChunked>();
 * \endcode
 */
class CadenaTransformacions : public Transformacio {
protected:
    typedef std::shared_ptr<Transformacio> transformacio_ptr;
public:
    CadenaTransformacions() {
        // buit
    }
    CadenaTransformacions(const CadenaTransformacions & ct)
        : cadena_(ct.cadena_)
    {
        // buit
    }
    CadenaTransformacions& operator=(const CadenaTransformacions & ct) {
        cadena_ = ct.cadena_;
        return *this;
    }

    template<typename TTransformacio>
    void afegeix() {
        transformacio_ptr trans(new TTransformacio());
        cadena_.push_back( trans );
    }
    virtual std::string transforma(const std::string & bloc) const;
    virtual std::string finalitza() const;
    virtual std::string destransforma(const std::string & text,
                                      EstatTransformacioInversa & estat) const;
    size_t pasos() const { return cadena_.size(); }
private:
    std::vector<transformacio_ptr> cadena_;
};

} // ns tfc

#endif // _TRANSFORMACIONS_H_

// vim:set ts=4 et ai: //
