#if !defined(_TFCWEB_URLV2_H_)
#define _TFCWEB_URLV2_H_

/*!
 * \file
 * \author Toni Corvera
 *
 * $Id: Uri.h 483 2013-06-17 17:41:04Z toni $
 */
/*
 * Copyright (c) 2012 Toni Corvera
 *
 * This file is part of TFCWeb.
 *
 * TFCWeb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TFCWeb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TFCWeb. If not, see <http://www.gnu.org/licenses/>.
 */

#include "Excepcions.h"
#include "portabilitat.h"

#include <ostream>
#include <stdexcept>
#include <string>

namespace tfc {

/*!
 * \brief Indica un error en l'anàlisi sintàctic d'una URL
 */
class ErrorAnalisiUrl : public ErrorTFC {
public:
    explicit ErrorAnalisiUrl(const std::string & w = "Error analitzant URL")
        : ErrorTFC(w)
    {
        // buit
    }
};

/*!
 * \brief Representació d'una URL en el domini del projecte
 * \sa http://tools.ietf.org/html/rfc2068#section-3.2
 */
class Uri {
public:    
    /*!
     * \param u URL a analitzar
     * \throws ErrorAnalisiUrl si \p u no es pot analitzar correctament
     * \invariant Url construida sense error <=> \p u és una URL vàlida
     */
    explicit Uri(const std::string & u) throw (ErrorAnalisiUrl);

    //! Constructor de còpia
    Uri(const Uri & u)
        : esquema_(u.esquema_),
          autoritat_(u.autoritat_),
          ruta_(u.ruta_),
          params_(u.params_),
          query_(u.query_),
          fragment_(u.fragment_)
    {
        // buit
    }

    //! Operador d'equivalència
    bool operator==(const Uri &) const;

    //! Operador d'assignació
    Uri& operator=(const Uri &);

    // Getters

    //! Protocol
    const std::string& esquema() const { return esquema_; }
    //! Host (equivalent a host() ) // TODO: Port?
    const std::string& autoritat() const { return autoritat_; }
    //! Host (equivalent a autoritat() ) \deprecated
    const std::string& host() const { return autoritat(); } // FIXME: host ==? autoritat
    //! Ruta
    const std::string& ruta() const { return ruta_; }
    //! Paràmetres \todo TODO: Gestionar o eliminar
    const std::string& params() const { return params_; }
    //! Consulta
    const std::string& query() const { return query_; }
    //! Fragment (marcador)
    const std::string& fragment() const { return fragment_; }

    // Setters

    //! Protocol
    void esquema(const std::string & v) { esquema_ = v; }
    //! Host (equivalent a host() )
    void autoritat(const std::string & v) { autoritat_ = v; }
    /*!
     * \brief Defineix la ruta
     * \throw ErrorAnalisiUrl Si no es pot interpretar \p v correctament
     */
    void ruta(const std::string & v) throw (ErrorAnalisiUrl);
    //! Paràmetres \todo TODO: Gestionar o eliminar
    void params(const std::string & p) { params_ = p; }
    //! Consulta
    void query(const std::string & v) { query_ = v; }
    //! Fragment (marcador)
    void fragment(const std::string & v) { fragment_ = v; }

    /*!
     * \brief Comprova si la URL és absoluta (conté esquema i host)
     * \invariant u.es_url_absoluta() == ( u.esquema() != string() && u.autoritat() != string() )
     */
    bool es_uri_absoluta() const;

    /*!
     * \brief Comprova si la ruta és absoluta
     * \remarks Una ruta buida és equivalent a "/" per tant, és absoluta, si cal
     *          distingir, comprovar si \ref ruta() és buida
     */
    bool te_ruta_absoluta() const;

    /*!
     * \brief Operador de conversió a cadena
     * \invariant \c string(Url(x)) == x # FIXME: Test case
     */
    operator std::string() const { return str(); }

    /*!
     * \brief Url completa
     */
    std::string str() const;

    /*!
     * \brief Obté una versió normalitzada de \p u
     * \post \code normalitza(u) == normalitza(u) == normalitza(normalitza(u)) \endcode
     * \post Sigui U = normalitza(u), es complirà:<br />
     *      \code
     *       U.esquema() == u.esquema() &&
     *       U.autoritat() == u.autoritat() &&
     *       U.query() == u.query() &&
     *       U.fragment() == u.fragment()
     *      \endcode
     * \post Si \c u.es_absoluta() llavors \c U.es_absoluta() (però no al revès)
     */
    static Uri normalitza(const Uri & u);

    /*!
     * \brief Descodificació de URLs.
     * \param[in] url Cadena de la URL a descodificar
     * \param[out] error Si la cadena conté caràcters codificats incorrectament es
     *                   possarà en \c false.
     * \remarks Només es descodifiquen els caracters codificats amb signe de percentatge
     *          seguit de dos dígits hexadecimals
     * \sa codifica()
     */
    static std::string descodifica(const std::string & url, bool & error) throw ();

    /*!
     * \brief Codificació de URLs.
     * \param s URL a codificar
     * \remarks Tota la cadena \p s es codificarà (p.e. "http://example.com" es
     *          transformarà en "http%3A%2F%2Fexample%2Ecom")
     * \sa descodifica()
     */ // TODO: Caràcters segurs
    static std::string codifica(const std::string & s) throw ();

    //! Normalització
    Uri normalitzada() const { return normalitza(*this); }

private:
    /*!
     * \brief Versió amb excepcions de \ref descodifica(const std::string&, bool&)
     * \throws ErrorAnalisiUrl Si es produeix en error en descodificar
     */
    static std::string descodifica(const std::string&) throw (ErrorAnalisiUrl);

    std::string esquema_,   //!< Esquema de la URL
                autoritat_, //!< "Autoritat" (nom de host)
                ruta_,      //!< Ruta
                params_,    //!< Paràmetres (...;[params[;params]])
                query_,     //!< Cadena de consulta (...?[consulta])
                fragment_;  //!< Fragment (...#[fragment])

    //! \brief Operador d'impressió
    friend std::ostream& operator<<(std::ostream & os, const Uri & u);
};

} // ns tfc

#endif // _TFCWEB_URLV2_H_

// vim:set ts=4 et ai: //
