#if !defined(_PORTABILITAT_H_)
#define _PORTABILITAT_H_

/*!
 * \file
 * \brief Macros i pragmas d'ajuda a l'ús dels diferents compiladors i
 *        sistemes operatius. A més, s'inclouen algunes capçaleres problemàtiques
 *        (p.e. que necessiten define's específics)
 *
 * \remarks No s'ha provat en Windows en mode UNICODE
 * \author Toni Corvera
 *
 * $Id: portabilitat.h 483 2013-06-17 17:41:04Z toni $
 */
/*
 * Copyright (c) 2012 Toni Corvera
 *
 * This file is part of TFCWeb.
 *
 * TFCWeb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TFCWeb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TFCWeb. If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef _WIN32
#   ifndef _WIN32_WINNT
#       define _WIN32_WINNT 0x0501 // Boost vol aquesta definició
#   endif
// Evita que es defineixin min() i max() com a macros
// http://www.devx.com/tips/Tip/14540
#   ifndef NOMINMAX
#      define NOMINMAX
#   endif
// Evita inclusions de capçaleres poc comunes (en particular, evita
// incloure capçaleres que poden ser conflictives, com ara winsock.h vs winsock2.h)
// http://support.microsoft.com/kb/166474
#   define WIN32_LEAN_AND_MEAN
#endif

//! Indica que una variable no s'utilitzarà
#if defined(__GNUC__) || defined(__clang__)
#   define UNUSED __attribute__((unused))
#elif defined(_MSC_VER)
// http://stackoverflow.com/questions/7090998/the-unused-macro
#   define UNUSED __pragma(warning(suppress:4100))
#else
#   define UNUSED /* buit */
#endif

// Desactivació de warnings en MSVC:
// 4290: avís que especificacions `throw (excepció)` s'ignoren
#if defined(_MSC_VER)
#   pragma warning(disable: 4290)
#endif

// En Windows, utilitzar configuració per defecte
#ifdef _WIN32
#   define HAVE_CONFIG_H
#endif
#ifdef HAVE_CONFIG_H
#   include "config.h"
#endif
#ifdef _WIN32
// Correcció de defines per Windows
#   undef HAVE_UNISTD_H
#   undef ZLIB_CONST // En la versió de Windows inclosa no té cap efecte (vegeu Transformacions.cc)
#endif

// Boost Filesystem ha canviat les interfícies recentment, per versions que encara no fan
// servir les noves per defecte cal definir això
#define BOOST_FILESYSTEM_VERSION 3

// Codis de sortida semi-estàndard
#if defined(HAVE_SYSEXITS_H) && !defined(WIN32)
#   include <sysexits.h>
#else
enum {
    EX_OK = 0,
    EX__BASE       = 64,
    EX_USAGE       = 64,
    EX_DATAERR     = 65,
    EX_NOINPUT     = 66,
    EX_NOUSER      = 67,
    EX_NOHOST      = 68,
    EX_UNAVAILABLE = 69,
    EX_SOFTWARE    = 70,
    EX_OSERR       = 71,
    EX_OSFILE      = 72,
    EX_CANTCREAT   = 73,
    EX_IOERR       = 74,
    EX_TEMPFAIL    = 75,
    EX_PROTOCOL    = 76,
    EX_NOPERM      = 77,
    EX_CONFIG      = 78,
    EX__MAX        = 78
};
#endif

#if defined(_WIN32)
#   define ZLIB_WINAPI
#   include "lib/zlib-1.2.5/zlib.h"
#else
#   include <zlib.h>
#endif

#define TFC_STRINGIFY_(s)    #s
// Indirecció necessaria per passar macros com argument
#define TFC_STRINGIFY(s)    TFC_STRINGIFY_(s)

// Informació de versions
// Compilador
#if defined(__clang__) // També defineix __GNUC__!
#   define TFC_COMPILADOR \
                        "clang++ "\
                        TFC_STRINGIFY(__clang_major__)"."\
                        TFC_STRINGIFY(__clang_minor__)"."\
                        TFC_STRINGIFY(__clang_patchlevel__)
#elif defined(__MINGW32__)
#   define TFC_COMPILADOR \
                        "MinGW "\
                        TFC_STRINGIFY(__MINGW32_MAJOR_VERSION)"."\
                        TFC_STRINGIFY(__MINGW32_MINOR_VERSION)", "\
                        "g++ "\
                        TFC_STRINGIFY(__GNUC__)"."\
                        TFC_STRINGIFY(__GNUC_MINOR__)"."\
                        TFC_STRINGIFY(__GNUC_PATCHLEVEL__)
#elif defined(__GNUC__)
#   define TFC_COMPILADOR \
                        "g++ "\
                        TFC_STRINGIFY(__GNUC__)"."\
                        TFC_STRINGIFY(__GNUC_MINOR__)"."\
                        TFC_STRINGIFY(__GNUC_PATCHLEVEL__)
#elif defined(_MSC_VER)
/* // Versió numèrica:
  Major: (_MSC_FULL_VER / 1000000)
  Minor: ((_MSC_FULL_VER % 1000000) / 10000)
  Build: (_MSC_FULL_VER % 10000)
*/
#   if _MSC_VER == 1700
#       define TFC_COMPILADOR "Visual C++ 11"
#   elif _MSC_VER == 1600
#       define TFC_COMPILADOR "Visual C++ 2010"
#   else
#       define TFC_COMPILADOR "Visual C++"
#   endif
#else
#   define TFC_COMPILADOR   "compilador no suportat"
#endif
// Arquitectura
#if defined(__amd64__) || defined(_M_AMD64)
#   define TFC_ARQUITECTURA   "amd64"
#elif defined(__i386__) || defined(__i386) || defined(_M_IX86)
#   define TFC_ARQUITECTURA   "x86"
#else
#   define TFC_ARQUITECTURA   "plataforma no suportada"
#endif
// Sistema operatiu
#if defined(__linux__)
#   define TFC_SO   "Linux"
#elif defined(_WIN64)
#   define TFC_SO   "Windows x86_64"
#elif defined(_WIN32)
#   define TFC_SO   "Windows x86"
#else
#   define TFC_SO   "sistema operatiu no suportat"
#endif
// Boost
#include <boost/version.hpp>
#if defined(BOOST_VERSION)
#   define  TFC_BOOST \
                        "Boost "\
                        BOOST_LIB_VERSION
/* // El preprocessador no fa els càlculs
                        TFC_STRINGIFY((BOOST_VERSION / 100000))"."\
                        TFC_STRINGIFY((BOOST_VERSION / 100) % 1000)"."\
                        TFC_STRINGIFY(BOOST_VERSION % 100)
*/
#else
#   define  TFC_BOOST   "?"
#endif

// Versió mínima de Boost suportada (principalment per Boost Filesystem)
// 1.44: Nova interfície disponible
// 1.48: Introducció de canonical()
#if BOOST_VERSION < 104800
#   error Cal versió 1.48 o superior de Boost
#endif

// Inclussions necessàries per les interfícies de sota
#include <fstream>
#include <memory>
#include <string>
#include <stdexcept>
#include <boost/filesystem.hpp>

namespace tfc {

/*!
 * \brief Eines per treballar de manera similar en diferents plataformes
 */
namespace portabilitat {

/*!
 * \brief Inicialitza característiques pròpies de la plataforma
 */
void init();

// Problemes:
// * getenv i familia poden no ser segures en programes multi-fil
// * la modificació de l'entorn no és portable
// Vegeu:
// - En GNU Lib C getenv i familia són segures
//   segons http://www.delorie.com/gnu/docs/glibc/libc_554.html
// - En Windows CRT no ho són
//   segons http://msdn.microsoft.com/en-us/library/tehxacec.aspx
//   es proveuen getenv_s i _putenv_s que són més segures però tampoc
//   es poden fer servir en multi-fil
//   http://msdn.microsoft.com/en-us/library/tb2sfw2z.aspx

/*!
 * \brief Obté el valor d'una variable d'entorn
 * \note Versió portable i \em thread-safe de \c getenv
 * \param var Nom de la variable
 * \param[out] existeix (opcional, \c NULL per ignorar) es possarà en \c true
 *             si la variable està definida
 * \return Valor de la variable. Cadena buida si la variable no està definida.
 */
std::string getenv(const std::string & var, bool * existeix = 0);

/*!
 * \brief Defineix el valor d'una variable d'entorn
 * \note Versió portable i \em thread-safe de \c setenv
 * \param var Nom de la variable
 * \param valor Valor de la variable
 * \return Si s'ha pogut assignar el valor correctament
 */
bool setenv(const std::string & var, const std::string & valor);

/*!
 * \brief Elimina una variable d'entorn
 * \note Versió portable i \em thread-safe de \c unsetenv
 * \param var Nom de la variable
 * \return Si s'ha pogut eliminar correctament
 */
bool unsetenv(const std::string & var);

/*!
 * \brief Obté la ruta (directori) del sistema per als fitxers temporals
 */
std::string ruta_temporal() throw (std::runtime_error);

/*!
 * \brief Creació d'un fitxer temporal
 * \param patro Patró de text a utilitzar per anomenar el fitxer
 * \param[out] nom Nom del fitxer creat
 * \param dir Directori en què crear el fitxer temporal
 * \return El fitxer obert per a escriptura, en mode binari
 * \post L'\c ofstream retornat s'ha obert sense error
 */
std::shared_ptr<std::ofstream> fitxer_temporal(const std::string & patro,
                                               std::string & nom,
                                               const std::string & dir = ruta_temporal())
                                                    throw (std::runtime_error);


/*!
 * \brief Comprova si el procés actual té permís d'execució sobre el fitxer
 *        apuntat per \p f
 * \pre \p f és un fitxer regular
 */
bool permis_execucio(const boost::filesystem::path & f);

} // ns portabilitat

} // ns tfc

#endif // _PORTABILITAT_H_

// vim:set ts=4 et ai: //
