#if !defined(_UTILS_H_)
#define _UTILS_H_

/*!
 * \file
 * \brief Funcions de suport vàries
 *
 * \author Toni Corvera
 *
 * $Id: utils.h 483 2013-06-17 17:41:04Z toni $
 */
/*
 * Copyright (c) 2012 Toni Corvera
 *
 * This file is part of TFCWeb.
 *
 * TFCWeb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TFCWeb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TFCWeb. If not, see <http://www.gnu.org/licenses/>.
 */

#include "portabilitat.h"

#include <istream>
#include <sstream>
#include <string>
#include <ostream>
#include <boost/bind.hpp>
#include <boost/date_time.hpp>
#include <boost/filesystem.hpp>
#include <boost/function.hpp>

namespace tfc {

// Forward declarations
template<typename T, typename TDesti> class Te_Operador_Insercio;
template<typename T, typename TOrigen> class Te_Operador_Extraccio;

/*!
 * \brief Funcions i algorismes de suport
 */
namespace utils {

//! \brief Tipus dels manipuladors de \c std::ios_base
// Definició en forma de signatura C:
typedef std::ios_base&(manipulador_ios_base_t)(std::ios_base&);
//! \brief Tipus compatible de manipuladors de \c std::ostream
// Definició amb boost::function:
typedef boost::function<std::ostream& (std::ostream&)> manipulador_ostream_t;

} // ns utils

} // ns tfc

/*!
 * \brief Inserció de manipuladors d'ostream amb signatures basades en \c boost::function
 * \internal
 * Encara que la signatura de \ref manipulador_ostream_t és compatible amb
 * els manipuladors d'ostream's, cal definir l'operador d'inserció que els executi,
 * altrament s'imprimirà l'objecte "boost::function" corresponent (produïnt la cadena "1")
 */
std::ostream& operator<<(std::ostream & os, tfc::utils::manipulador_ostream_t fn);

namespace tfc {

namespace utils {

/*!
 * \brief Obté la representació en forma de cadena d'un objecte
 * \note L'objecte ha de ser imprimible per a convertir-lo
 * \note Operació inversa de \ref from_str()
 * \tparam T Tipus de l'objecte a representar
 * \param obj Objecte a representar
 * \param base Base numèrica a aplicar
 * \returns Representació textual de \a obj
 * \remarks \c boost::lexical_cast és preferible allà on es pugui fer servir,
 *          l'avantatge d'aquesta funció és poder indicar una base numèrica
 *          diferent de la decimal
 * \sa from_string()
 * \pre L'operador d'inserció (\c <<) de \p T sobre una \c std::istream és definit
 */
template<typename T> std::string to_string(const T & obj,
                                           manipulador_ios_base_t base = std::dec)
{
    Te_Operador_Insercio<T, std::ostringstream>();
    std::ostringstream ss;
    ss << base << obj;
    return ss.str();
}

/*!
 * \brief Extreu un valor de la cadena \p s
 * \tparam T Tipus de l'objecte a extreure
 * \param s Valor d'entrada
 * \param base Base numèrica a aplicar
 * \returns Valor extret
 * \note Operació inversa de \ref to_string()
 * \remarks Per base decimal \c boost::lexical_cast és preferible, però no
 *          no permet indicar altres bases
 * \remarks \c std::ios_base&(*)(std::ios_base&) és el tipus dels modificadors de base
 *          numèrica de la biblioteca estàndard
 *
 * Exemple: \code
 * int n = from_string<int>("0x0f", std::hex);
 * assert( n == 15 );
 * \endcode
 * \pre L'operador d'extracció (\c >>) de \p T sobre una \c std::ostream és definit
 */
template<typename T> T from_string(const std::string & s,
                                   manipulador_ios_base_t base = std::dec)
{
    Te_Operador_Extraccio<T, std::istringstream>();
    T ret;
    std::istringstream ss(s);
    ss >> base >> ret;
    return ret;
}

/*!
 * \brief Obté el darrer element d'una col·lecció.
 * Exemple: \code
 * char c = final(std::string("Prova")); // c == 'a'
 * \endcode
 * \pre s.size() > 0
 */ // FIXME: Massa general? Usar basic_string? http://www.sgi.com/tech/stl/basic_string.html
template<typename Str> const typename Str::value_type final(const Str & s) {
    assert( s.size() > 0 ); // PRECONDICIÓ
    return *(--s.end());
}

/*!
 * \brief Comprova si un fitxer o directori és llegible
 * \param p Ruta a comprovar
 * \pre \a p existeix
 * \pre \a p és un fitxer o directori
 * \remarks Les precondicions són relaxades:
 *          Si no es respecten símplement retornarà \c false
 */
bool es_llegible(const boost::filesystem::path & p) throw ();

/*!
 * \brief Comprova si una ruta és canònica
 * \param p Ruta a comprovar
 * \pre \a p existeix
 */
bool es_canonic(const boost::filesystem::path & p) throw ();

/*!
 * \brief Amplia una cadena per l'esquerra
 * \param s Cadena a ampliar
 * \param mida Mida mínima de la cadena
 * \param f Caràcter amb què ampliar-la
 */
std::string lpad(const std::string & s, size_t mida, char f = ' ');

/*!
 * \brief Amplia una cadena per la dreta
 * \param s Cadena a ampliar
 * \param mida Mida mínima de la cadena
 * \param f Caràcter amb què ampliar-la
 */
std::string rpad(const std::string & s, size_t mida, char f = ' ');

/*!
 * \brief Formata una hora/data
 * \param t Marca de temps
 * \param fmt Format (\em flags de format de Boost Date Time, pràcticament iguals
 *            a \c strftime())
 * \returns Data formatada
 * \sa http://www.boost.org/doc/libs/1_49_0/doc/html/date_time/date_time_io.html#date_time.format_flags
 * \details
 * Exemple: Data HTTP actual: \code
 * std::time_t ara_t = second_clock::universal_time();
 * boost::posix_time::ptime ara = boost::posix_time::from_time_t(ara_t);
 * std::string d1 = formata_data(ara, "%a, %d %b %Y %T GMT");
 * // o
 * std::string d2 = formata_data(ara_t, "%a, %d %b %Y %T GMT");
 * assert( d1 == d2 );
 * \endcode
 */
std::string formata_data(const boost::posix_time::ptime & t, const std::string & fmt);

// Nota: La interpretació de dates textuals no s'ha generalitzat, principalment
// perquè Boost Date Time no inclou lectura de fusos horaris estàndards
// vegeu llegeix_data_http() per el cas concret del format utilitzat en HTTP

/*!
 * \copydoc formata_data(const boost::posix_time::ptime&, const std::string&)
 */
std::string formata_data(const std::time_t t, const std::string & fmt);

//! Detalls d'implementació
namespace impl {

// Cal una implementació amb nom no polimòrfic per què el compilador
// la interpreti correctament en les funcions generadores que utilitzen
// boost::bind (vegeu les utils::each_shl())
// XXX: parametritzar Separador potser és excessiu
template<typename Iterador, typename Separador = std::string>
std::ostream& each_shl(std::ostream& os,
                       Iterador ini, Iterador fi,
                       const Separador & sep)
{
   while (ini != (fi-1)) { // Fins el penúltim element, l'imprimim seguit del separador
       os << *ini << sep;
       ini++;
   }
   return os << *ini; // Últim element, sense separador
}

} // ns impl

/*!
 * \brief Impressió de contenidors.
 * \param os Destinació
 * \param ini Principi de la iteració
 * \param fi Final de la iteració
 * \param sep Separador. Cadena inserida després de cada element excepte el darrer
 * \tparam Iterador tipus d'iterador
 * \tparam Separador tipus del separador
 * \details
 * Aplica l'operador \c &lt;&lt; a tots els elements entre \p ini i \p fi.
 * <br />
 * Exemple:
 * \code
 * vector<...> v;
 * cout << each_shl(cout, v.begin(), v.end()) << endl;
 * \endcode
 */
template<typename Iterador, typename Separador = std::string>
std::ostream& each_shl(std::ostream& os,
                       Iterador ini, Iterador fi,
                       const Separador & sep = "\n") {
    return impl::each_shl(os, ini, fi, sep);
}

/*!
 * \brief Impressió de contenidors amb destí implícit.
 * \param ini Principi de la iteració
 * \param fi Final de la iteració
 * \param sep Separador. Cadena inserida després de cada element excepte el darrer
 * \tparam Iterador tipus d'iterador
 * \tparam Separador tipus del separador

 * \details
 * Funció generadora. La inserció en un ostream el defineix com a destí.
 * Aplica l'operador \c &lt;&lt; a tots els elements entre \p ini i \p fi
 * <br />
 * Exemple:
 * \code
 * vector<...> v;
 * cout << each_shl(v.begin(), v.end());
 * \endcode
 */
template<typename Iterador, typename Separador = std::string>
manipulador_ostream_t each_shl(Iterador ini, Iterador fi, const Separador & sep = "\n") {
    // No compila: return boost::bind<std::ostream&>(&each_shl<Iterador>, _1, ini, fi);
    return boost::bind<std::ostream&>(&impl::each_shl<Iterador>, _1, ini, fi, sep);
}

/*!
 * \brief Impressió de contenidors amb destí implícit.
 * \param c Contenidor
 * \param fi Final de la iteració
 * \param sep Separador. Cadena inserida després de cada element excepte el darrer
 * \tparam Contenidor tipus contenidor
 * \tparam Separador tipus del separador
 * \details
 * Funció generadora. La inserció en un ostream el defineix com a destí.
 * Aplica l'operador \c &lt;&lt; a tots els elements de la iteració \c c.begin()
 * a \c c.end() (equivalent a \code each_shl(c.begin(), c.end()) \endcode)
 * <br />
 * Exemple:
 * \code
 * vector<...> v;
 * v.push_back("hola");
 * v.push_back("món");
 * cout << each_shl(v) << endl;
 * \endcode
 * Generarà: \code
 * hola
 * món
 * \endcode
 */
template<typename Contenidor, typename Separador = std::string>
manipulador_ostream_t each_shl(const Contenidor & c, const Separador & sep = "\n") {
    return each_shl(c.begin(), c.end(), sep);
    // typedef typename Contenidor::const_iterator const_iterator;
    // return boost::bind<std::ostream&>(&impl::each_shl<const_iterator>, _1, c.begin(), c.end());
}


/*!
 * \brief Formata la hora/data actual
 * \param fmt Format (\em flags de format de Boost Date Time, pràcticament iguals
 *            a \c strftime())
 * \returns Data formatada
 * \sa http://www.boost.org/doc/libs/1_49_0/doc/html/date_time/date_time_io.html#date_time.format_flags
 * \sa formata_data(const std::time_t, const std::string&)
 * \sa formata_data(const boost::posix_time::ptime &, const std::string&)
 */
inline std::string formata_data(const std::string & fmt) {
    return formata_data(boost::posix_time::second_clock::universal_time(), fmt);
}

} // ns tfc::utils

// Constraints
//
// La tècnica de les constraints està explicada pel propi Bjarne Stroustrup a:
// <http://www2.research.att.com/~bs/bs_faq2.html#constraints>
//
// Són construccions evaluables en la compilació que impossen restriccions en
// els tipus de les plantilles (prevenint la compilació si els tipus passats no
// compleixen la condició impossada). Un cop evaluades no generen codi.
//

/*!
 * \ingroup Constraints
 * \brief Restricció (\em constraint): \p T respon a la interfície de
 *        \c std::pair<std::string, std::string>
 * \tparam T Tipus sobre el que aplicar la restricció
 * \details
 * Exemple: \code
 * template<class Iterador> void imprimeix_primers(Iterador a, Iterador b) {
 *     Es_Pair_De_Strings<typename Iterador::value_type>();
 *     for (Iterador i=a; i!=b; ++i) {
 *         std::cout << i->first << std::endl;
 *     }
 * }
 * std::vector< std::pair< std::string, std::string > > vps;
 * f(vps.begin());
 * \endcode
 */
template<typename T>
class Es_Pair_De_Strings {
    static void constraint(T & t) {
        const std::string first = t.first;
        const std::string second = t.second;
    }
public:
    Es_Pair_De_Strings() {
        UNUSED void (*p_ct)(T&) = constraint;
    }
};

/*!
 * \ingroup Constraints
 * \brief Restricció (\em constraint): \p T es pot cridar com si fós una
 *        funció amb la signatura <tt>void (void)</tt>
 * \tparam T Tipus d'objecte sobre el que aplicar la restricció
 * \details
 * Exemple: \code
 * struct fnobj {
 *     void operator()() {}
 * };
 * void fn() {}
 * Es_Pot_Cridar_Sense_Arguments<fnobj>(); // OK
 * Es_Pot_Cridar_Sense_Arguments<void (*)()>(); // OK
 * Es_Pot_Cridar_Sense_Arguments<decltype(fn)>(); // OK (C++11)
 * \endcode
 */
template<typename T>
class Es_Pot_Cridar_Sense_Arguments {
    static void constraint(T & t) {
        t();
    }
public:
    Es_Pot_Cridar_Sense_Arguments() {
        UNUSED void (*p_ct)(T&) = constraint;
    }
};

/*!
 * \ingroup Constraints
 * \brief Restricció (\em constraint): \p T té definit l'operador d'inserció
 *        (\c <<) sobre objectes de tipus \p TDesti
 * \tparam T Tipus d'objecte sobre el que aplicar la restricció
 * \tparam TDesti Tipus d'objecte en el que insertar \p T
 * \details
 * Exemple: \code
 * struct S {};
 * std::ostream& operator<<(std::ostream& s, const S &) {
 *     return s;
 * }
 * Te_Operador_Insercio<S>(); // OK
 * \endcode
 */
template<typename T, typename TDesti = std::ostream>
class Te_Operador_Insercio {
    static void constraint(T & t, TDesti & desti) {
        desti << t;
    }
public:
    Te_Operador_Insercio() {
        UNUSED void (*p_ct)(T&, TDesti&) = constraint;
    }
};

/*!
 * \ingroup Constraints
 * \brief Restricció (\em constraint): \p T té definit l'operador
 *        d'extracció (\c >>) sobre objectes del tipus \p TOrigen
 * \tparam T Tipus d'objecte sobre el que aplicar la restricció
 * \tparam TOrigen Tipus desde el que es pot extreure \p T
 * \details
 * Exemple: \code
 * struct S {};
 * std::istream& operator>>(std::istream& s, const S &) {
 *     return s;
 * }
 * Te_Operador_Extraccio<S>(); // OK
 * \endcode
 */
template<typename T, typename TOrigen = std::istream>
class Te_Operador_Extraccio {
    static void constraint(T & t, TOrigen & origen) {
        origen >> t;
    }
public:
    Te_Operador_Extraccio() {
        UNUSED void (*p_ct)(T&, TOrigen&) = constraint;
    }
};

} // ns tfc

#endif // _UTILS_H_

// vim:set ts=4 et ai: //
