
/*!
 * \file
 * \brief Tests de tfc::Uri.
 * \sa tfc::Uri
 *
 * \author Toni Corvera
 *
 * $Id: TestUri.cc 483 2013-06-17 17:41:04Z toni $
 */
/*
 * Copyright (c) 2012 Toni Corvera
 *
 * This file is part of TFCWeb.
 *
 * TFCWeb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TFCWeb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TFCWeb. If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <limits>
#include <sstream>
#include <string>

#include <Uri.h>

#include "comu.h"

using namespace std;
using namespace tfc;

namespace tfc {

namespace Fixtures {

struct FxtUri {

    FxtUri()
        : str_www_google_com("www.google.com"), // ambigu!
          www_google_com("www.google.com"), // ambigu!
          str_http_www_google_com("http://www.google.com"),
          http_www_google_com("http://www.google.com"),
          str_https_www_google_com_query("https://www.google.com/search?q=c%2B%2B"),
          https_www_google_com_query("https://www.google.com/search?q=c%2B%2B"),
          index_html("index.html"),
          str_barra_barra_str("//str"), // s'entendrà com a http://str/ !
          barra_barra_str("//str")
    {
        // buit
    }

    const string str_www_google_com,
                 str_http_www_google_com,
                 str_https_www_google_com_query,
                 str_barra_barra_str;

    const Uri www_google_com,
              http_www_google_com,
              https_www_google_com_query,
              index_html,
              barra_barra_str;
};

} // ns Fixtures

} // ns tfc

BOOST_FIXTURE_TEST_SUITE( Uri_TestSuite, tfc::Fixtures::FxtUri )

/*!
 * Dreçera per provar tfc::utils::normalitza()
 */
inline string norm(const string & p) {
    const Uri u(p);
    return string(u.normalitzada());
}

BOOST_AUTO_TEST_CASE( construccio )
{
    BOOST_CHECK_NO_THROW( Uri("../index.html") );
    BOOST_CHECK_NO_THROW( Uri("././../index.html?q=home#frg") );

    // TODO: Realment és incorrecta sobre l'RFC?, sembla que no
    //BOOST_CHECK_THROW( Uri("http://http://www.google.com"), tfc::ErrorAnalisiUrl );
}

BOOST_AUTO_TEST_CASE( igualtat )
{
    BOOST_CHECK_EQUAL( Uri(str_www_google_com), Uri(str_www_google_com) );
    BOOST_CHECK_EQUAL( Uri(str_https_www_google_com_query), Uri(str_https_www_google_com_query) );
    BOOST_CHECK_EQUAL( Uri(barra_barra_str), Uri(barra_barra_str) );
}

BOOST_AUTO_TEST_CASE( assignacio )
{
    Uri c(https_www_google_com_query);
    BOOST_CHECK_EQUAL( c, https_www_google_com_query );
    c = www_google_com;
    BOOST_CHECK_EQUAL( c, www_google_com );
    c = Uri(".");
    BOOST_CHECK_EQUAL( c, Uri(".") );
}

BOOST_AUTO_TEST_CASE( copia )
{
    Uri c_google(www_google_com), c_barres(barra_barra_str);
    BOOST_CHECK_EQUAL( www_google_com, c_google );
    BOOST_CHECK_EQUAL( barra_barra_str, c_barres );
}

BOOST_AUTO_TEST_CASE( absolutes_i_relatives )
{
    Uri abs("http://www.example.com"),
            rel("../index.html"),
            ruta_abs("/index.html");
    // TODO: Escriure tests
}

BOOST_AUTO_TEST_CASE( postcondicio_normalitzacio )
{
    const Uri & u = https_www_google_com_query;
    const Uri nu1(Uri::normalitza(u));
    const Uri nu2(Uri::normalitza(u));
    const Uri nunu(Uri::normalitza(nu1));
    // POST1 ( normalitza(u) == normalitza(u) == normalitza(normalitza(u)) )
    BOOST_CHECK_EQUAL( nu1, nu2 );
    BOOST_CHECK_EQUAL( nu1, nunu );
    // POST2
    BOOST_CHECK_EQUAL( nu1.esquema(), u.esquema() );
    BOOST_CHECK_EQUAL( nu1.autoritat(), u.autoritat() );
    BOOST_CHECK_EQUAL( nu1.query(), u.query() );
    BOOST_CHECK_EQUAL( nu1.fragment(), u.fragment() );
}

BOOST_AUTO_TEST_CASE( normalitzacio_completes )
{
    BOOST_CHECK_EQUAL( norm("http://www.google.com/.."),
                            "http://www.google.com/");
}

BOOST_AUTO_TEST_CASE( normalitzacio_parcials_absolutes )
{
    BOOST_CHECK_EQUAL( norm("/"), "/" );
    // TODO: Escriure nous tests
#if 0 // TODO: Les dobles barres tenen un significat especial!
    BOOST_CHECK_EQUAL( norm("//"), "/" );
    BOOST_CHECK_EQUAL( norm("//.//"), "/" );
    BOOST_CHECK_EQUAL( norm("///"), "/" );
    BOOST_CHECK_EQUAL( norm("/index.html"), "/index.html");
    BOOST_CHECK_EQUAL( norm("//index.html"), "/index.html");
#endif
}

BOOST_AUTO_TEST_CASE( normalitzacio_parcials_relatives )
{
    BOOST_CHECK_EQUAL( norm("."), "" );
    BOOST_CHECK_EQUAL( norm("./.."), "" );
    BOOST_CHECK_EQUAL( norm("../index.html"), "/index.html");
    BOOST_CHECK_EQUAL( norm("dir/../index.html"), "/index.html");
    // FIXME: és implícit '/' ó './' ?
    BOOST_CHECK_EQUAL( norm("index.html"), "/index.html");
}

inline void compara_string(const string & s) {
    BOOST_CHECK_EQUAL( string(Uri(s)), s );
}

BOOST_AUTO_TEST_CASE( parse )
{
    //BOOST_CHECK_THROW(frase_estat(indefinit), ErrorTipusPeticioHTTPDesconeguda);

    compara_string( str_www_google_com );
    compara_string( str_https_www_google_com_query );
    compara_string( str_barra_barra_str );
}


BOOST_AUTO_TEST_CASE( descodificacio )
{
    bool error;
    const string r = Uri::descodifica("%2F", error);
    BOOST_CHECK_EQUAL( r, "/" );
    // Exemple usat en la documentació de la capçalera Url.h
    BOOST_CHECK_EQUAL( "http://example.com",
                        Uri::descodifica("http%3A%2F%2Fexample%2Ecom",error) );
}

BOOST_AUTO_TEST_CASE( codificacio )
{
    const string r = Uri::codifica("/");
    BOOST_CHECK_EQUAL( r, "%2F" );
    // Exemple usat en la documentació de la capçalera Url.h
    BOOST_CHECK_EQUAL( "http%3A%2F%2Fexample%2Ecom",
                        Uri::codifica("http://example.com") );
}

BOOST_AUTO_TEST_SUITE_END()

// vim:set ts=4 et ai: //
